/*
 * gtktube: a gtk frontend for freetv like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gtktube.h"
#include "gtktube_search.h"
#include "freetv_search.h"
#include "gtk_utils.h"
#include "utils.h"
#include "marshal.h"


#define DEBUG_XML 0



enum
{
  SIGNAL_SEARCH_PROGRESS_STATUS,
  SIGNAL_SEARCH_FINISHED,
  SIGNAL_LAST
};

static gint freetv_search_signals[SIGNAL_LAST];
static char *freetv_playlist_urls[] = {
   "http://mafreebox.freebox.fr/freeboxtv/playlist.m3u",
   "http://rootsrv.privfbx.proxad.net/pub/playlist.m3u",
};


static void
freetv_search_parse_m3u_response(FreeTvSearch *search, SoupMessage *msg)
{
  const char *data = msg->response_body->data;
  gint datalen = msg->response_body->length;
  gboolean is_m3u = FALSE;
  const char *s, *p, *e;
  char *url;
  gboolean stream_is_not_a_video = FALSE;
  Video *video;

  if (datalen < 128)
    return;

  if (strncmp(data, "#EXTM3U", strlen("#EXTM3U")))
    return;

  video = video_new();

  /* Read line by line */
  s = p = data;
  while (*p)
   {
     if (*p != '\n')
      {
	p++;
	continue;
      }

     e = p;

     /* string to parse is between s and e */
     if (strncmp(s, "#EXTM3U", strlen("#EXTM3U")) == 0)
       is_m3u = TRUE;
     else if (strncmp(s, "#EXTINF:", strlen("#EXTINF:")) == 0)
      {
	/* #EXTINF:0,2 - France 2 HD */
	/* #EXTINF:0,10085 - Monte-Carlo Doualiya */
	while (s < e)
	 {
	   s++;
	   if (s[0] == ' ' && s[-1] == '-' && s[-2] == ' ')
	    {
	      s++;
	      video_set_title(video, s, e-s);
	    }
	 }
      }
     else if (strncmp(s, "#EXTVLCOPT:", strlen("#EXTVLCOPT:")) == 0)
      {
	/* #EXTVLCOPT:no-video */
	if (strncmp(s+strlen("#EXTVLCOPT:"), "no-video", strlen("no-video")) == 0)
	  stream_is_not_a_video = TRUE;
	/* #EXTVLCOPT:ts-es-id-pid */
	/* #EXTVLCOPT:audio-track-id=1005 */

      }
     else if (strncmp(s, "rtsp://", strlen("rtsp://")) == 0)
      {
	/* rtsp://mafreebox.freebox.fr/fbxtv_pub/stream?namespace=1&service=201&flavour=hd */

	/* Register the stream if it's a video */
	if (stream_is_not_a_video == FALSE)
	 {
	   url = g_strndup(s, e-s);
	   video_set_webpage(video, url);
	   g_free(url);
	   search->videos = g_list_prepend(search->videos, video);
	   video = video_new();
	 }
	stream_is_not_a_video = FALSE;
      }

     s = e+1;
     p++;
   }

  g_object_unref(video);
 
}




/*
 *
 * Networks layers to make request
 *
 *
 *
 */
static void
freetv_search_async_callback(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
  FreeTvSearch *search = user_data;

  trace("\n");

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code))
   {
     /* FreeTv return error in the body */
     search->http_return_code = msg->status_code;
   } 
  else
   {
     search->http_return_code = 0;
   }

  trace("download status code: %d\n", msg->status_code);
  trace("%d bytes downloaded\n", (gsize)msg->response_body->length);
  //trace("%s\n--\n", msg->response_body->data);

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code) == FALSE)
    freetv_search_parse_m3u_response(search, msg);

  gdk_threads_enter();
  g_signal_emit(search, freetv_search_signals[SIGNAL_SEARCH_FINISHED], 0);
  gdk_threads_leave();
}

static void
freetv_search_got_chunk_cb(SoupMessage *msg, SoupBuffer *chunk, gpointer user_data)
{
  FreeTvSearch *search = user_data;
  guint64 content_length;
  guint64 received_content_length;

  trace("\n");

  if (msg->response_body == NULL)
    return;

  content_length = soup_message_headers_get_content_length(msg->response_headers);
  received_content_length = msg->response_body->length;
  trace("got = %lld/%lld\n", received_content_length, content_length);

  gdk_threads_enter();
  g_signal_emit(search, freetv_search_signals[SIGNAL_SEARCH_PROGRESS_STATUS], 0, content_length, received_content_length);
  gdk_threads_leave();
}

/*
 * At date of Jan2010, we have a bug with the freebox Tv service,
 * the server didn't response at http://mafreebox.freebox.fr/freeboxtv/playlist.m3u
 * and reset the connection. libsoup restart all messages automaticaly, so
 * we need to trigger this, and try the second url... if working
 *
 * If we are no url, cancel the message
 * Else, requeue the message, after changing the url
 *
 */
static void
freetv_search_restarted_cb(SoupMessage *msg, gpointer user_data)
{
  FreeTvSearch *search = user_data;
  SoupURI *uri;

  trace("\n");

  search->retry_for_this_url++;
  if (search->retry_for_this_url < 1)
    return;

  search->retry_for_this_url = 0;
  search->current_url_index++;
  if (search->current_url_index >= G_N_ELEMENTS(freetv_playlist_urls))
   {
     soup_session_cancel_message(search->async_http_session, msg, SOUP_STATUS_CANT_CONNECT);
     return;
   }

  uri = soup_uri_new(freetv_playlist_urls[search->current_url_index]);
  soup_message_set_uri(msg, uri);
  soup_uri_free(uri);
  soup_session_requeue_message(search->async_http_session, msg);
}



static gboolean
freetv_load_playlist_by_url(FreeTvSearch *search, const char *url)
{
  SoupMessage *msg;

  trace("url=%s\n", url);

  if (search->msg)
   {
     g_warning("Clearing the previous FreeTv query ...\n");
     g_object_unref(search->msg);
     search->msg = NULL;
   }

  msg = soup_message_new("GET", url);
  if (msg == NULL)
    return FALSE;

  g_signal_connect(G_OBJECT(msg), "got-chunk", G_CALLBACK(freetv_search_got_chunk_cb), search);
  g_signal_connect(G_OBJECT(msg), "restarted", G_CALLBACK(freetv_search_restarted_cb), search);

  if (search->async_http_session == NULL)
    search->async_http_session = soup_session_async_new();

  soup_session_queue_message(search->async_http_session, msg, freetv_search_async_callback, search);

  return TRUE;
}






/*
 *
 * Public interface
 *
 *
 */




FreeTvSearch *
freetv_search_new(void)
{
  FreeTvSearch *search = g_object_new(FREETV_SEARCH_TYPE, NULL);

  return search;
}

gboolean
freetv_load_playlist(FreeTvSearch *search)
{
  trace("Loading playlist ...\n");

  search->retry_for_this_url = 0;
  search->current_url_index = 0;

  return freetv_load_playlist_by_url(search, freetv_playlist_urls[0]);
}

void
freetv_search_set_http_session(FreeTvSearch *search, SoupSession *http_session)
{
  search->async_http_session = g_object_ref(http_session);
}

GList *
freetv_search_get_results(FreeTvSearch *search)
{
  GList *videos = search->videos;

  return videos;
}


/***********************************************************************
 *
 *
 *
 *
 * GObject things
 *
 *
 *
 *
 **********************************************************************/
G_DEFINE_TYPE(FreeTvSearch, freetv_search, G_TYPE_OBJECT)

static void
freetv_search_finalize(GObject *object)
{
  FreeTvSearch *search = (FreeTvSearch *)(object);

  if (search->async_http_session)
    g_object_unref(search->async_http_session);
  search->async_http_session = NULL;

  if (search->msg)
    g_object_unref(search->msg);

  G_OBJECT_CLASS (freetv_search_parent_class)->finalize(object);
}

static void
freetv_search_class_init(FreeTvSearchClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

  gobject_class->finalize = freetv_search_finalize;

  freetv_search_signals[SIGNAL_SEARCH_FINISHED] =
        g_signal_new("search-finished",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

  freetv_search_signals[SIGNAL_SEARCH_PROGRESS_STATUS] =
        g_signal_new("progress-status",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     gtktube_marshal_VOID__UINT64_UINT64,
                     G_TYPE_NONE,
		     2,
		     G_TYPE_UINT64,
		     G_TYPE_UINT64);

}

static void
freetv_search_init(FreeTvSearch *search)
{
  trace("\n");

  search->async_http_session = NULL;
}




/***********************************************************************
 *
 *
 *
 *
 * Plugin interface
 *
 *
 *
 *
 **********************************************************************/

static void
on_freetv_search_finished(FreeTvSearch *search, gpointer user_data)
{
  struct gtktube_search_plugin *priv = user_data;

  if (priv->search_finished_cb)
   {
     GList *videos = freetv_search_get_results(search);
     priv->search_finished_cb(videos, priv->search_finished_cb_priv_data);
   }

}

static void
on_freetv_progress_status(FreeTvSearch *search, guint64 content_length, guint64 received_content_length, gpointer user_data)
{
  struct gtktube_search_plugin *priv = user_data;

  if (priv->progress_status_cb)
    priv->progress_status_cb(content_length, received_content_length, priv->progress_status_cb_priv_data);

}


struct gtktube_search_plugin plugin_freetv =
{
   .name           = "freetv",
   .description    = "FreeTv Videos",

   .type           = GTKTUBE_PLUGIN_TV,
   .logo           = "free.png",

   //.have_search    = TRUE,

   .set_http_session = (SearchPluginSetHttpSessionCallback)freetv_search_set_http_session,
   .load_playlist = (SearchPluginLoadPlaylistCallback)freetv_load_playlist,
};


static int
freetv_plugin_init(void)
{
  trace("\n");

  FreeTvSearch *priv  = freetv_search_new();
  plugin_freetv.priv = priv;

  g_signal_connect(priv, "search-finished", G_CALLBACK(on_freetv_search_finished), &plugin_freetv);
  g_signal_connect(priv, "progress-status", G_CALLBACK(on_freetv_progress_status), &plugin_freetv);

  return gtktube_search_plugin_register(&plugin_freetv);
}

static void
freetv_plugin_exit(void)
{
  trace("\n");
  gtktube_search_plugin_unregister(&plugin_freetv);
}



GTKTUBE_PLUGIN_DEFINE(freetv,
    		      "FreeTv Video Service",
		      "Luc Saillard <luc@saillard.org>",
		      PACKAGE_VERSION,
		      freetv_plugin_init,
		      freetv_plugin_exit)


