/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>


#include "page_intro.h"
#include "page_view.h"

#include "youtube_search.h"
#include "youtube_suggest.h"

#include "gtktube_plugin.h"
#include "gtktube_search.h"
#include "gtktube.h"
#include "gtk_utils.h"
#include "utils.h"


static GtkWidget *gtktube_create_main_window(struct gtktube_private *priv);

static int   opt_verbose;
static int   opt_console;
static char *opt_search_string;
#if HAVE_JSONGLIB
static char *opt_suggest_string;
#endif
       int   opt_no_video;


static GOptionEntry options_entries[] =
{
  { "console", 'c', 0, G_OPTION_ARG_NONE, &opt_console, N_("Print result on the terminal (DEBUG ONLY)"), NULL},
  { "verbose", 'v', 0, G_OPTION_ARG_NONE, &opt_verbose, N_("Show more output on the console"), NULL},
  { "search", 's', 0, G_OPTION_ARG_STRING, &opt_search_string, N_("Search KEYWORD on youtube"), "KEYWORD"},
#if HAVE_JSONGLIB
  { "suggest", 'S', 0, G_OPTION_ARG_STRING, &opt_suggest_string, N_("Suggest search with KEYWORD"), "KEYWORD"},
#endif
  { "no-video", 0, 0, G_OPTION_ARG_NONE, &opt_no_video, N_("Disable video"), NULL},
  { NULL }
};





int main(int argc, char **argv)
{
  GtkWidget *main_window;
  GOptionContext *option_context;
  GError *gerror = NULL;
  struct gtktube_private *priv;


  setlocale(LC_ALL, "");

  bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
  textdomain(GETTEXT_PACKAGE);

  g_thread_init(NULL);
  g_type_init();

  option_context = g_option_context_new(_("- Gtk Frontend for Youtube service"));
  g_option_context_add_main_entries(option_context, options_entries, PACKAGE_NAME);
  g_option_context_add_group(option_context, gtk_get_option_group(TRUE));
  g_option_context_parse(option_context, &argc, &argv, &gerror);

  priv = g_new0(struct gtktube_private, 1);

  priv->async_http_session = soup_session_async_new();

  if (opt_verbose)
   {
     SoupLogger *logger;

     logger = soup_logger_new(SOUP_LOGGER_LOG_HEADERS, -1);
     soup_session_add_feature(priv->async_http_session, SOUP_SESSION_FEATURE(logger));
     g_object_unref(logger);
   }

  if (ENABLE_DEBUG && opt_console && opt_search_string)
   {
     GMainLoop *loop = g_main_loop_new (NULL, TRUE);
     void on_debug_search_finished(YoutubeSearch *search, gpointer user_data)
      {
	GMainLoop *loop = user_data;
	GList *videos = youtube_search_get_results(search);
	trace("videos = %p\n", videos);
	while (videos)
	 {
	   Video *video = videos->data;
	   trace("video %s\n", video->webpage);
	   trace(" title: %s\n", video->title);
	   trace(" author: %s\n", video->author_name);
	   trace(" description: %s\n", video->description);
	   trace(" stats: views=%d favorites=%d\n", video->views, video->favorites);
	   trace(" duration: %ds\n", video->duration);
	   trace("\n");
	   videos = videos->next;
	 }
	g_main_loop_quit(loop);
      };

     YoutubeSearch *search = youtube_search_new();
     youtube_search_set_http_session(search, priv->async_http_session);
     youtube_search_search_with_keywords(search, opt_search_string);
     g_signal_connect(search, "search-finished", G_CALLBACK(on_debug_search_finished), loop);

     g_main_loop_run(loop);
     exit(0);
   }

#if HAVE_JSONGLIB
  if (ENABLE_DEBUG && opt_console && opt_suggest_string)
   {
     GMainLoop *loop = g_main_loop_new (NULL, TRUE);
     void on_debug_search_finished(YoutubeSuggest *search, gpointer user_data)
      {
	GMainLoop *loop = user_data;
	GList *list = youtube_suggest_get_results(search);
	while (list)
	 {
	   trace(" %s\n", (char *)list->data);
	   list = list->next;
	 }
	g_main_loop_quit(loop);
	trace("--\n");
      };

     YoutubeSuggest *search = youtube_suggest_new();
     youtube_suggest_set_http_session(search, priv->async_http_session);
     youtube_suggest_search_with_keywords(search, opt_suggest_string);
     g_signal_connect(search, "search-finished", G_CALLBACK(on_debug_search_finished), loop);

     g_main_loop_run(loop);
     exit(0);
   }
#endif

  gst_init(&argc, &argv);

  gtk_init(&argc, &argv);


  /* Load plugins */
  gtktube_plugin_init();

  /* Set default plugin */
  priv->search_plugin = gtktube_search_plugin_get_first();

  main_window = gtktube_create_main_window(priv);
  gtk_widget_show(main_window);
  gtk_widget_grab_focus(priv->search_entry);

  if (opt_search_string)
   {
     gtk_entry_set_text(GTK_ENTRY(priv->search_entry), opt_search_string);
     g_signal_emit_by_name(priv->search_entry, "activate", NULL);
   }
  gtk_main();

  gtktube_plugin_exit();

  return 0;
}





static gboolean
on_close_window_event(GtkWidget *widget, GdkEvent  *event, gpointer data)
{
  gtk_main_quit();
  return FALSE;
}





static void
menu_quit(GtkAction *action)
{
  gtk_main_quit();
}

static void
menu_not_implemented(GtkAction *action)
{
  trace("\n");
}

static void
menu_toggle_fullscreen(GtkAction *action, gpointer user_data)
{
  trace("\n");

}

static void
menu_show_about_dialog(GtkAction *action, struct gtktube_private *priv)
{
  const gchar *authors[] = {
    "Luc Saillard",
    NULL
  };
  gtk_show_about_dialog (GTK_WINDOW(priv->main_window),
                         "name", "GtkTube",
                         "version", PACKAGE_VERSION,
                         "copyright", "(C) Luc Saillard",
                         "comments", "Show Youtube like sites.",
                         "authors", authors,
                         NULL);
}




static GtkActionEntry actions_entries[] = {
  { "PlaylistMenu", NULL, N_("_Playlist") },
  { "VideoMenu",    NULL, N_("_Video") },
  { "HelpMenu",     NULL, N_("_Help") },

  { "Quit", GTK_STOCK_QUIT,
    N_("_Quit"), "<control>Q",
    N_("Quit"),
    G_CALLBACK(menu_quit) },

  { "About", GTK_STOCK_ABOUT,                  /* name, stock id */
    N_("_About"), "<control>A",                /* label, accelerator */
    N_("About"),                               /* tooltip */
    G_CALLBACK(menu_show_about_dialog) },


  /* Playlist Menu */
  { "Remove", GTK_STOCK_REMOVE,
    N_("_Remove this playlist"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

  { "Up", GTK_STOCK_GO_UP,
    N_("Move _Up"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

  { "Down", GTK_STOCK_GO_DOWN,
    N_("Move _Down"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

  /* Video Menu */
  { "Play", GTK_STOCK_MEDIA_PLAY,
    N_("_Play"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

  { "Pause", GTK_STOCK_MEDIA_PAUSE,
    N_("_Pause"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

  { "Skip", GTK_STOCK_MEDIA_NEXT,
    N_("_Skip"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

  { "GoYoutube", GTK_STOCK_JUMP_TO,
    N_("_Go to Youtube"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented) },

};

/* http://library.gnome.org/devel/gtk/stable/GtkActionGroup.html#GtkToggleActionEntry */
static GtkToggleActionEntry toggle_actions_entries[] = {
  /* Video Menu */
  { "CompactMode", NULL,
    N_("_Compact Mode"), NULL,
    NULL,
    G_CALLBACK(menu_not_implemented),
    FALSE },


  { "FullScreen", NULL,
    N_("_FullScreen"), "F11",
    NULL,
    G_CALLBACK(menu_toggle_fullscreen),
    FALSE },


};

static const gchar *ui_info =
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='PlaylistMenu'>"
"      <menuitem action='Remove'/>"
"      <separator/>"
"      <menuitem action='Up'/>"
"      <menuitem action='Down'/>"
"      <separator/>"
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='VideoMenu'>"
"      <menuitem action='Play'/>"
"      <menuitem action='Pause'/>"
"      <menuitem action='Skip'/>"
"      <separator/>"
"      <menuitem action='GoYoutube'/>"
"      <separator/>"
"      <menuitem action='CompactMode'/>"
"      <menuitem action='FullScreen'/>"
"    </menu>"
"    <menu action='HelpMenu'>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"</ui>";










/*
 *
 * .---------------------------------------------------------------------.
 * | Menu | Video |                                               | Help |
 * |---------------------------------------------------------------------|
 * |                                                                     |
 * |   .-----------.                                                     |
 * |   |           |                                                     |
 * |   |           |         [Search tags__________]                     |
 * |   |           |           .........                                 |
 * |   |           |           ..........                                |
 * |   '-----------'                                                     |
 * |                                                                     |
 * |---------------------------------------------------------------------|
 * | Please wait while downloading the movie ....                        |
 * '---------------------------------------------------------------------'
 *
 * .---------------------------------------------------------------------.
 * | Menu | Video |                                               | Help |
 * |---------------------------------------------------------------------|
 * | .------.           |                                                |
 * | |      | ****      |                                                |
 * | |______| video1    |                                                |
 * |____________________|                                                |
 * | .------.           |                                                |
 * | |      | **        |                                                |
 * | |______| video2    |                                                |
 * |____________________|                                                |
 * |                    |                                                |
 * |---------------------------------------------------------------------|
 * | Please wait while downloading the movie ....                        |
 * '---------------------------------------------------------------------'
 *
 */

static GtkWidget *
gtktube_create_main_window(struct gtktube_private *priv)
{
  GtkWidget *window, *vbox;
  GtkWidget *menubar;
  GtkActionGroup *actions;
  GError *gerror = NULL;


  priv->main_window = window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  g_signal_connect(G_OBJECT(window), "delete_event", G_CALLBACK(on_close_window_event), NULL);

  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
  gtk_window_set_icon_name(GTK_WINDOW(window), "preferences-system-network");
  gtk_container_set_border_width(GTK_CONTAINER(window), 10);
  gtk_window_set_title(GTK_WINDOW(window), _("GtkTube - Web videos sites frontend"));
  gtk_widget_set_size_request(window, 740, 440);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);

  priv->ui = gtk_ui_manager_new();

  actions = gtk_action_group_new("Actions");
  gtk_action_group_set_translation_domain(actions, GETTEXT_PACKAGE);
  gtk_action_group_add_actions(actions, actions_entries, G_N_ELEMENTS(actions_entries), priv);
  gtk_action_group_add_toggle_actions(actions, toggle_actions_entries, G_N_ELEMENTS(toggle_actions_entries), priv);


  gtk_ui_manager_insert_action_group(priv->ui, actions, 0);
  gtk_window_add_accel_group(GTK_WINDOW(window), gtk_ui_manager_get_accel_group(priv->ui));
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);

  if (gtk_ui_manager_add_ui_from_string(priv->ui, ui_info, -1, &gerror) == FALSE)
   {
     error("building menus failed: %s", gerror->message);
     g_error_free(gerror);
     return NULL;
   }

  menubar = gtk_ui_manager_get_widget(priv->ui, "/MenuBar");
  gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);

  priv->notebook = gtk_notebook_new();
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(priv->notebook), FALSE);
  gtk_box_pack_start(GTK_BOX(vbox), priv->notebook, TRUE, TRUE, 0);

  priv->statusbar = gtk_statusbar_new();
  gtk_box_pack_end(GTK_BOX(vbox), priv->statusbar, FALSE, FALSE, 0);

  gtk_widget_show_all(window);

  /* Attach all pages to the notebook */
  gtk_notebook_append_page(GTK_NOTEBOOK(priv->notebook), page_intro_new(priv), NULL);
  gtk_notebook_append_page(GTK_NOTEBOOK(priv->notebook), page_view_new(priv), NULL);

  return window;
}








void
gtktube_show_page(struct gtktube_private *priv, guint page_index)
{
  gtk_notebook_set_current_page(GTK_NOTEBOOK(priv->notebook), page_index);
}



