/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gtktube.h"
#include "gtk_utils.h"
#include "utils.h"

#include "page_view.h"

/*
 *
 *
 *
 */
static void
on_page_intro_search_finished(GList *videos, gpointer user_data)
{
  struct gtktube_private *priv = user_data;

  trace("\n");

  if (videos == NULL)
   {
     error("Display error for no result ...\n");
     return;
   }

  if (priv->videos)
   {
     g_list_foreach(priv->videos, (GFunc)g_object_unref, NULL);
     g_list_free(priv->videos);
   }

  priv->videos = videos;

  gtktube_show_page(priv, INDEX_PAGE_VIEW);
  page_view_update_search(priv);

}

static void
on_page_intro_progress_status(guint64 content_length, guint64 received_content_length, gpointer user_data)
{
  struct gtktube_private *priv = user_data;

  trace("\n");

#if GTK_CHECK_VERSION(2,16,0)
  if (content_length == 0)
   {
     /* Only update the progress bar */
     gtk_entry_progress_pulse(GTK_ENTRY(priv->search_entry));
   }
  else
   {
     gdouble progress = ((gdouble)content_length / received_content_length) * 100.0;
     if (progress > 100)
       progress = 100;

     gtk_entry_set_progress_fraction(GTK_ENTRY(priv->search_entry), progress);
   }
#endif
}


/*
 * We got results to suggest some query for the user.
 *
 */
#if HAVE_JSONGLIB
static void
page_intro_query_for_suggests_got_results(YoutubeSuggest *search, gpointer user_data)
{
  GList *list;
  GtkTreeIter iter;
  struct gtktube_private *priv = user_data;

  trace("\n");

  gtk_list_store_clear(priv->suggests_store);

  /* Synchronize the filtered model. */
  gtk_entry_completion_complete(priv->search_completion);

  list = youtube_suggest_get_results(search);

  while (list)
   {
     gtk_list_store_append(priv->suggests_store, &iter);
     gtk_list_store_set(priv->suggests_store, &iter, 0, list->data, -1);

     list = list->next;
   }

  g_list_foreach(list, (GFunc)g_free, NULL);
  g_list_free(list);

  g_object_unref(priv->search_suggest);
  priv->search_suggest = NULL;
}
#endif


/*
 * Query the web, to find completion for this entry
 */
#if HAVE_JSONGLIB
static void
page_intro_query_for_suggests(struct gtktube_private *priv)
{
  const gchar *prefix;

  if (priv->search_suggest)
   {
     trace("A search is already in progress.. so do nothing\n");
     return;
   }

  prefix = gtk_entry_get_text(GTK_ENTRY(priv->search_entry));
  if (strlen(prefix) < 3)
    return;

  priv->search_suggest = youtube_suggest_new();
  youtube_suggest_set_http_session(priv->search_suggest, priv->async_http_session);
  youtube_suggest_search_with_keywords(priv->search_suggest, prefix);
  g_signal_connect(priv->search_suggest, "search-finished", G_CALLBACK(page_intro_query_for_suggests_got_results), priv);

}
#endif


/*
 * Called when the search entry is changed.
 * We must allow the "Watch" button, if they are some characters
 */
static void
page_intro_entry_changed(GtkWidget *entry, gpointer user_data)
{
  const char *text;
  struct gtktube_private *priv = user_data;

  trace("\n");

  text = gtk_entry_get_text(GTK_ENTRY(entry));
  if (strlen(text) == 0)
   {
     gtk_widget_set_sensitive(priv->search_button, FALSE);
     return;
   }

  gtk_widget_set_sensitive(priv->search_button, TRUE);

#if HAVE_JSONGLIB
  /* Launch a query ... */
  if (GTK_WIDGET_REALIZED(GTK_WIDGET(entry)) && priv->search_plugin->have_suggest_completion)
    page_intro_query_for_suggests(priv);
#endif
}

static void
page_intro_entry_activate(GtkWidget *entry, gpointer user_data)
{
  struct gtktube_private *priv = user_data;
  const char *text = NULL;
  trace("\n");

  if (entry)
    text = gtk_entry_get_text(GTK_ENTRY(entry));

  gtktube_search_plugin_set_http_session(priv->search_plugin, priv->async_http_session);
  gtktube_search_plugin_set_search_finished_cb(priv->search_plugin, on_page_intro_search_finished, priv);
  gtktube_search_plugin_set_progress_status_cb(priv->search_plugin, on_page_intro_progress_status, priv);

  /* TODO: cancel any search in progress */
  gtktube_search_plugin_start_new_search(priv->search_plugin, text);

  /* while the search is in progress, cancel button, and entry */
  gtk_widget_set_sensitive(priv->search_button, FALSE);
  gtk_widget_set_sensitive(priv->search_entry, FALSE);
}

static void
page_intro_load_playlist(struct gtktube_private *priv)
{
  gtktube_search_plugin_set_http_session(priv->search_plugin, priv->async_http_session);
  gtktube_search_plugin_set_search_finished_cb(priv->search_plugin, on_page_intro_search_finished, priv);
  gtktube_search_plugin_set_progress_status_cb(priv->search_plugin, on_page_intro_progress_status, priv);
  gtktube_search_plugin_load_playlist(priv->search_plugin);
}



static void
page_intro_button_watch_clicked(GtkWidget *button, gpointer user_data)
{
  GtkWidget *entry = user_data;

  trace("\n");
  if (entry)
    g_signal_emit_by_name(entry, "activate", NULL);
}

#if 0
static gboolean
suggest_completion_match_cb(GtkEntryCompletion * completion, const gchar *key, GtkTreeIter *iter, gpointer user_data)
{
    return TRUE;
}
#endif

#if 0
static gboolean
suggest_completion_match_selected(GtkEntryCompletion *completion, GtkTreeModel *model, GtkTreeIter *iter)
{
  gchar *str;
  GtkWidget *entry;

  entry = gtk_entry_completion_get_entry(completion);
  gtk_tree_model_get(GTK_TREE_MODEL(model), iter, 1, &str, -1);
  gtk_entry_set_text(GTK_ENTRY(entry), str);
  gtk_editable_set_position(GTK_EDITABLE (entry), -1);
  g_free (str);

  return TRUE;
}
#endif


/*
 *
 * .---------------------------------------------------------------------.
 * | Menu | Video |                                               | Help |
 * |---------------------------------------------------------------------|
 * |                                                                     |
 * |   .-----------.         GTKTUBE                                     |
 * |   |           |                                                     |
 * |   |           |         [Search tags__________]                     |
 * |   |           |           .........                                 |
 * |   |           |           ..........                                |
 * |   '-----------'                                                     |
 * | {change provider}                                                   |
 * |---------------------------------------------------------------------|
 * | Please wait while downloading the movie ....                        |
 * '---------------------------------------------------------------------'
 *
 */

GtkWidget *
page_intro_new(struct gtktube_private *priv)
{
  GtkWidget *hbox, *hbox2, *vbox, *vbox2, *image, *label, *entry, *button;
  GdkPixbuf *logo;
  PangoAttrList  *attrs_label;
  PangoAttribute *attribute;

  g_return_val_if_fail(priv->search_plugin != NULL, NULL);

  /* Cleanup if we have initialized a previous page */
  if (priv->suggests_store)
   {
     gtk_list_store_clear(priv->suggests_store);
     g_object_unref(priv->suggests_store);
     priv->suggests_store = NULL;
   }

  if (priv->search_completion)
   {
     g_object_unref(priv->search_completion);
     priv->search_completion = NULL;
   }

  /* Create the object to do the suggest search */
  if (priv->search_plugin->have_suggest_completion)
   {
     priv->suggests_store = gtk_list_store_new(1, G_TYPE_STRING);

     priv->search_completion = gtk_entry_completion_new();
     gtk_entry_completion_set_model(priv->search_completion, GTK_TREE_MODEL(priv->suggests_store));
     gtk_entry_completion_set_minimum_key_length(priv->search_completion, 3);
     gtk_entry_completion_set_popup_single_match(priv->search_completion, FALSE);

     gtk_entry_completion_set_text_column(priv->search_completion, 0);
   }


  hbox = gtk_hbox_new(FALSE, 0);

  /*
   *
   *    .--------------.
   *    |              |
   *    |              |
   *    |              |
   *    |              |
   *    '______________'
   *
   *
   */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 4);


  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, FALSE, 4);

  if (priv->search_plugin->logo)
    logo = gtk_utils_get_pixbuf_from_icon_name(priv->search_plugin->logo, 128);
  else
    logo = gtk_utils_get_pixbuf_from_icon_name("video-display", 128);
  image = gtk_image_new_from_pixbuf(logo);
  gtk_box_pack_start(GTK_BOX(vbox2), image, FALSE, FALSE, 4);
  g_object_unref(logo);
  gtk_widget_set_size_request(GTK_WIDGET(image), 128+64, -1);

  button = gtk_button_new_with_mnemonic(_("_Change provider"));
  gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 4);
  gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
  gtk_button_set_focus_on_click(GTK_BUTTON(button), FALSE);



  /*
   *
   *          GtkTube
   *
   *          [Search tags__________]
   *            .........
   *            ..........
   *
   *
   */

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 4);

  /* The " GtkTube " label in big bold face */
  attrs_label = pango_attr_list_new ();

  attribute = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
  attribute->start_index = 0;
  attribute->end_index = -1;
  pango_attr_list_insert(attrs_label, attribute);

  attribute = pango_attr_size_new(24 * PANGO_SCALE);
  attribute->start_index = 0;
  attribute->end_index = -1;
  pango_attr_list_insert(attrs_label, attribute);

  label = gtk_label_new("GtkTube");
  gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
  gtk_label_set_attributes(GTK_LABEL(label), attrs_label);
  gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, FALSE, 4);

  pango_attr_list_unref(attrs_label);

  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, TRUE, 4);

  /* The GtkEntry to enter the search */
  label = gtk_label_new("Enter a keyword to start watching videos.");
  gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 4);

  hbox2 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 4);


  priv->search_entry = entry = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox2), entry, TRUE, TRUE, 4);
  gtk_entry_set_max_length(GTK_ENTRY(entry), 64);
  gtk_entry_set_completion(GTK_ENTRY(entry), priv->search_completion);
#if GTK_CHECK_VERSION(2,16,0)
  gtk_entry_set_icon_from_stock(GTK_ENTRY(entry), GTK_ENTRY_ICON_PRIMARY, GTK_STOCK_FIND);
  gtk_entry_set_icon_activatable(GTK_ENTRY(entry), GTK_ENTRY_ICON_PRIMARY, FALSE);
#endif

  if (priv->search_plugin->have_search)
   {
     /* Watch button, if use doesn't known how to press enter */
     priv->search_button = button = gtk_button_new_with_mnemonic(_("_Watch"));
     gtk_box_pack_start(GTK_BOX(hbox2), button, FALSE, FALSE, 4);
     gtk_widget_set_sensitive(button, FALSE);
     gtk_button_set_focus_on_click(GTK_BUTTON(button), FALSE);
     g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(page_intro_button_watch_clicked), entry);

     /* Connect other signals */
     g_signal_connect(G_OBJECT(entry), "changed", G_CALLBACK(page_intro_entry_changed), priv);
     g_signal_connect(G_OBJECT(entry), "activate", G_CALLBACK(page_intro_entry_activate), priv);
   }
  else if (priv->search_plugin->load_playlist)
   {
     /* TV plugin doesn't have a search, so load the list of stream directly */
     page_intro_load_playlist(priv);
   }

  gtk_widget_show_all(hbox);
  return hbox;
}

