/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gtktube.h"
#include "gtk_utils.h"
#include "video.h"
#include "video_view.h"
#include "utils.h"

static void page_view_on_state_changed(VideoView *view, int state, struct gtktube_private *priv);

static gboolean page_view_video_timer_update(struct gtktube_private *priv);










static gboolean
page_view_update_thumbnails(struct gtktube_private *priv)
{
  trace("\n");
  g_list_foreach(priv->videos, (GFunc)video_update_thumbnail, priv->async_http_session);

  return TRUE;
}

static void
page_view_button_media_play_pause_clicked(GtkWidget *button, struct gtktube_private *priv)
{
  trace("\n");

  if (priv->current_video_selected == NULL)
    return;

  if (priv->video_state == GST_STATE_PLAYING)
   {
     video_view_pause(priv->video_widget);

     g_return_if_fail(priv->update_timer_id > 0);

     g_source_remove(priv->update_timer_id);
     priv->update_timer_id = 0;
   }
  else if (priv->video_widget->pipeline)
   {
     video_view_play(priv->video_widget);

     g_return_if_fail(priv->update_timer_id == 0);
     priv->update_timer_id = g_timeout_add_seconds(1, (GSourceFunc)page_view_video_timer_update, priv);
   }
  else
   {
     /* We need to create the pipeline and play the video */
     video_view_set_video(priv->video_widget, priv->current_video_selected, priv->async_http_session);

     g_return_if_fail(priv->update_timer_id == 0);
     priv->update_timer_id = g_timeout_add_seconds(1, (GSourceFunc)page_view_video_timer_update, priv);
   }
}

static void
page_view_button_media_stop_clicked(GtkWidget *button, struct gtktube_private *priv)
{
  trace("\n");

  if (priv->current_video_selected == NULL)
    return;

  video_view_stop(priv->video_widget);
  page_view_on_state_changed(priv->video_widget, GST_STATE_NULL, priv);
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(priv->progress_bar), NULL);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(priv->progress_bar), 1);
  priv->video_state = GST_STATE_NULL;
  if (priv->update_timer_id)
   {
     g_source_remove(priv->update_timer_id);
     priv->update_timer_id = 0;
   }
}

static gboolean
page_view_on_progress_bar_clicked(GtkWidget *widget, GdkEventButton *event, struct gtktube_private *priv)
{
  int x, y;
  gdouble percent;

  trace("\n");

  if (widget->allocation.width == 0)
    return TRUE;

  if (priv->video_widget->pipeline == NULL)
    return TRUE;

  gtk_widget_get_pointer(widget, &x, &y);

  percent = (gdouble)x/widget->allocation.width;

  trace("seeking to %d%%\n", (guint)(percent*100));

  gst_element_seek_simple(priv->video_widget->pipeline, GST_FORMAT_PERCENT, GST_SEEK_FLAG_FLUSH, (guint64)(percent*100));

  return TRUE;
}


static void
page_view_on_end_of_stream_cb(VideoView *view, struct gtktube_private *priv)
{
  trace("\n");
  page_view_button_media_stop_clicked(NULL, priv);
}

static void
page_view_on_buffering_info(VideoView *view, int percent, struct gtktube_private *priv)
{
  char *text;

  if (priv->progress_bar == NULL)
    return;
  
  trace("\n");

  text = g_strdup_printf(_("Buffering: %d%%"), percent);

  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(priv->progress_bar), (gdouble)percent/100);
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(priv->progress_bar), text);

  g_free(text);
}


static void
page_view_on_state_changed(VideoView *view, int state, struct gtktube_private *priv)
{
  const char *stock;

  if (state == priv->video_state)
    return;

  priv->video_state = state;

  if (state == GST_STATE_PAUSED || state == GST_STATE_NULL)
    stock = GTK_STOCK_MEDIA_PLAY;
  else if (state == GST_STATE_PLAYING)
    stock = GTK_STOCK_MEDIA_PAUSE;
  else
    return;

  gtk_image_set_from_stock(GTK_IMAGE(priv->play_pause_image), stock, GTK_ICON_SIZE_BUTTON);
}

static gboolean
page_view_video_timer_update(struct gtktube_private *priv)
{
  gint64 position, duration;
  char *text = NULL;
  double percent = 0;

  if (priv->progress_bar == NULL)
    return TRUE;

  if (priv->video_state != GST_STATE_PLAYING)
    return TRUE;

  /* call me again */
  video_view_get_times(VIDEO_VIEW(priv->video_widget), &position, &duration);

  if (position >=0 && duration >= 0)
   {
     text= g_strdup_printf("%u:%02u:%02u / %u:%02u:%02u", 
	 (guint) (((GstClockTime)(position)) / (GST_SECOND * 60 * 60)),
	 (guint) ((((GstClockTime)(position)) / (GST_SECOND * 60)) % 60),
         (guint) ((((GstClockTime)(position)) / GST_SECOND) % 60),
	 (guint) (((GstClockTime)(duration)) / (GST_SECOND * 60 * 60)),
	 (guint) ((((GstClockTime)(duration)) / (GST_SECOND * 60)) % 60),
         (guint) ((((GstClockTime)(duration)) / GST_SECOND) % 60));
	 
     percent = (double)position/duration;
   }
  else if (position >= 0)
   {
     text= g_strdup_printf("%" GST_TIME_FORMAT , GST_TIME_ARGS(position));
   }
  else
    return TRUE;

  if (percent > 1.0)
    percent = 1.0;

  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(priv->progress_bar), percent);
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(priv->progress_bar), text);

  g_free(text);

  return TRUE;
}


static gboolean
page_view_video_selected(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  struct gtktube_private *priv = user_data;
  Video *video = g_object_get_data(G_OBJECT(widget), "video");

  trace("\n");

  /* Stop the current video if necessary */
  page_view_button_media_stop_clicked(NULL, priv);

  /* Reset all widgets state to unselected */
  g_list_foreach(priv->videos, (GFunc)video_set_state, GINT_TO_POINTER(FALSE));
  video_set_state(video, TRUE);

  /* ... and play the next video */
  priv->current_video_selected = video;
  page_view_button_media_play_pause_clicked(NULL, priv);

  return TRUE;
}

static gboolean
page_view_go_to_home(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  struct gtktube_private *priv = user_data;

  trace("\n");

  /* Stop the current video if necessary */
  page_view_button_media_stop_clicked(NULL, priv);

  gtktube_show_page(priv, INDEX_PAGE_SEARCH);

  return TRUE;
}


















/*
 *
 *
 */
void
page_view_update_search(struct gtktube_private *priv)
{
  GList *iter;

  priv->current_video_selected = NULL;
  gtk_container_foreach(GTK_CONTAINER(priv->videos_box), (GtkCallback)gtk_widget_destroy, NULL);

  iter = priv->videos;
  while (iter)
   {
     Video *video = iter->data;
     GtkWidget *widget_video;

     widget_video = video_create_widget(video);
     if (widget_video)
      {
	g_object_set_data(G_OBJECT(widget_video), "video", video);
	g_signal_connect(G_OBJECT(widget_video), "button-press-event", G_CALLBACK(page_view_video_selected), priv);
	gtk_box_pack_start(GTK_BOX(priv->videos_box), widget_video, FALSE, FALSE, 4);
	video_update_thumbnail(video, priv->async_http_session);
      }

     iter = iter->next;
   }

  if (priv->videos)
   {
     priv->current_video_selected = priv->videos->data;
     video_set_state(priv->videos->data, TRUE);
   }

  g_timeout_add_seconds(12, (GSourceFunc)page_view_update_thumbnails, priv);
}





/*
 * |---------------------------------------------------------------------|
 * | [<<][>][o][>>] [======================>                           ] |
 * |---------------------------------------------------------------------|
 */
static GtkWidget *
page_view_create_toolbar(struct gtktube_private *priv)
{
  GtkWidget *hbox, *button, *image, *pb, *evbox;

  hbox = gtk_hbox_new(FALSE, 0);

  button = gtk_button_new();
  image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PREVIOUS, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);

  button = gtk_button_new();
  priv->play_pause_image = image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(page_view_button_media_play_pause_clicked), priv);

  button = gtk_button_new();
  image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_STOP, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(page_view_button_media_stop_clicked), priv);

  button = gtk_button_new();
  image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_NEXT, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);


  priv->progress_bar = pb = gtk_progress_bar_new();
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(pb), _("Please wait"));
  gtk_widget_set_events(pb, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  g_signal_connect(G_OBJECT(pb), "button-press-event", G_CALLBACK(page_view_on_progress_bar_clicked), priv);

  evbox = gtk_event_box_new();
  gtk_container_add(GTK_CONTAINER(evbox), pb);
  gtk_box_pack_start(GTK_BOX(hbox), evbox, TRUE, TRUE, 12);


  return hbox;
}


/*
 * ,------------------------,
 * [<< ( Sort by date  ) >> |
 * '---( Sort by pop   )----'
 *     ( Sort by views )
 *     `---------------'
 */
static GtkWidget *
page_view_create_navbar(struct gtktube_private *priv)
{
  GtkWidget *hbox, *combo, *button, *image;

  hbox = gtk_hbox_new(FALSE, 0);

  button = gtk_button_new();
  gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
  image = gtk_image_new_from_stock(GTK_STOCK_HOME, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);
  g_signal_connect(G_OBJECT(button), "button-press-event", G_CALLBACK(page_view_go_to_home), priv);


  button = gtk_button_new();
  gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
  image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PREVIOUS, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);


  combo = gtk_combo_box_new_text();
  gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Sort by views"));
  gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Sort by date"));
  gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Sort by score"));
  gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
  gtk_box_pack_start(GTK_BOX(hbox), combo, TRUE, TRUE, 4);


  button = gtk_button_new();
  gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
  image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_NEXT, GTK_ICON_SIZE_BUTTON);
  gtk_button_set_image(GTK_BUTTON(button), image);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 2);


  return hbox;
}


/*
 * ,------------------------,
 * | .------.           | ^ |
 * | |      | ****      | | |
 * | |______| video1    | | |
 * |____________________| | |
 * | .------.           | | |
 * | |      | **        | | |
 * | |______| video2    | | |
 * |____________________| | |
 * |                    | v |
 * '------------------------'
 */
static GtkWidget *
page_view_create_search_result_view(struct gtktube_private *priv)
{
  GtkWidget *vbox, *sc;

  sc = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sc), GTK_SHADOW_NONE);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sc), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sc), vbox);

  priv->videos_box = vbox;

  return sc;
}


/*
 * |---------------------------------------------------------------------|
 * | .------.           |                                                |
 * | |      | ****      |                                                |
 * | |______| video1    |                                                |
 * |____________________|                                                |
 * | .------.           |                                                |
 * | |      | **        |                                                |
 * | |______| video2    |                                                |
 * |____________________|                                                |
 * |                    |                                                |
 * |---------------------------------------------------------------------|
 */
static GtkWidget *
page_view_create_main_view(struct gtktube_private *priv)
{
  GtkWidget *hbox, *vbox, *video_widget, *box;

  video_widget = video_view_new();
  priv->video_widget = (VideoView *)video_widget;
  g_signal_connect(G_OBJECT(video_widget), "buffering", G_CALLBACK(page_view_on_buffering_info), priv);
  g_signal_connect(G_OBJECT(video_widget), "end-of-stream", G_CALLBACK(page_view_on_end_of_stream_cb), priv);
  g_signal_connect(G_OBJECT(video_widget), "video-state-changed", G_CALLBACK(page_view_on_state_changed), priv);

  box = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box), video_widget, FALSE, FALSE, 4);
  //gtk_container_set_border_width(GTK_CONTAINER(align), 4);


  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), page_view_create_navbar(priv), FALSE, FALSE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), page_view_create_search_result_view(priv), TRUE, TRUE, 4);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 4);
  gtk_box_pack_start(GTK_BOX(hbox), box, FALSE, FALSE, 4);


  return hbox;
}


/*
 *
 * .---------------------------------------------------------------------.
 * | Menu | Video |                                               | Help |
 * |---------------------------------------------------------------------|
 * | .------.           |                                                |
 * | |      | ****      |                                                |
 * | |______| video1    |                                                |
 * |____________________|                                                |
 * | .------.           |                                                |
 * | |      | **        |                                                |
 * | |______| video2    |                                                |
 * |____________________|                                                |
 * |                    |                                                |
 * |---------------------------------------------------------------------|
 * |              [======================>                             ] |
 * |---------------------------------------------------------------------|
 * | Please wait while downloading the movie ....                        |
 * '---------------------------------------------------------------------'
 *
 *
 */

GtkWidget *
page_view_new(struct gtktube_private *priv)
{
  GtkWidget *main_view, *toolbar, *vbox_top;

  main_view = page_view_create_main_view(priv);
  toolbar = page_view_create_toolbar(priv);

  vbox_top = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox_top), main_view, TRUE, TRUE, 4);
  gtk_box_pack_start(GTK_BOX(vbox_top), toolbar, FALSE, FALSE, 4);

  gtk_widget_show_all(vbox_top);

  return vbox_top;
}

