/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include <config.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gtktube.h"
#include "gtk_utils.h"
#include "video.h"
#include "video_thumbnail.h"
#include "utils.h"



static GdkPixbuf *black_pixbuf;
static GdkPixbuf *gold_star_pixbuf;
static GdkPixbuf *grey_star_pixbuf;
static const GdkColor   color_when_selected = {0, 0x3000, 0x2000, 0xbf00};
static       GdkColor   color_normal;

enum
{
  SIGNAL_VIDEO_READY = 0,
  SIGNAL_LAST
};

static gint video_signals[SIGNAL_LAST];


/*
 * We have received a signal that tell us that a new thumbnail has been loaded,
 * so update the preview picture.
 */
static void
on_video_thumbnails_loaded(VideoThumbnail *thumbnail, gpointer user_data)
{
  Video *video = user_data;

  if (   (thumbnail->pixbuf != NULL)
      && (thumbnail->width == VIDEO_THUMBNAIL_WIDTH)
      && (thumbnail->height == VIDEO_THUMBNAIL_HEIGHT)
      && (video->preview_image != NULL))
   {
     gtk_image_set_from_pixbuf(GTK_IMAGE(video->preview_image), thumbnail->pixbuf);
   }

  video->thumbnail_is_loading = FALSE;
}



/*
 *
 *
 *
 */
void
video_set_webpage(Video *video, const char *href)
{
  video->webpage = g_strdup(href);
}

/*
 *
 *
 *
 */
void
video_set_author_name(Video *video, const char *text, gsize text_len)
{
  g_return_if_fail(text != NULL);
  g_return_if_fail(text_len > 0);
  video->author_name = g_strndup(text, text_len);
}

/*
 *
 *
 *
 */
void
video_set_title(Video *video, const char *text, gsize text_len)
{
  g_return_if_fail(text != NULL);
  g_return_if_fail(text_len > 0);
  video->title = g_strndup(text, text_len);
}

/*
 *
 *
 *
 */
void
video_set_description(Video *video, const char *text, gsize text_len)
{
  g_return_if_fail(text != NULL);
  g_return_if_fail(text_len > 0);
  video->description = g_strndup(text, text_len);
}

/*
 *
 *
 *
 */
void
video_set_duration(Video *video, int duration_in_sec)
{
  g_return_if_fail(duration_in_sec > 0);
  video->duration = duration_in_sec;
}


/*
 *
 *
 *
 */
void
video_set_published_date(Video *video, int since_epoch)
{
  video->date_published = since_epoch;
}

/*
 *
 *
 *
 */
void
video_set_statistics(Video *video, int views, int favorites)
{
  video->views = views;
  video->favorites = favorites;
}

void video_set_ratings(Video *video, gdouble avg, int min, int max, int num)
{
  video->avg_rating = avg;
  video->min_rating = min;
  video->max_rating = max;
  video->num_votes = num;
}

/*
 *
 *
 *
 */
void
video_add_thumbnail_uri(Video *video, const char *uri, int width, int height)
{
  VideoThumbnail *thumbnail = video_thumbnail_new(uri, width, height);
  g_signal_connect(thumbnail, "img-loaded", G_CALLBACK(on_video_thumbnails_loaded), video);

  if (width > VIDEO_THUMBNAIL_WIDTH || height > VIDEO_THUMBNAIL_HEIGHT)
    video->previews = g_list_prepend(video->previews, thumbnail);
  else
    video->thumbnails = g_list_prepend(video->thumbnails, thumbnail);
}


/*
 *
 *
 *
 */
#if 0
void
video_add_content(Video *video, const char *uri, const char *type, guint format)
{
  /* Only accept type flash */
  if (format != 5)
    return;

  if (g_strcmp0(type, "application/x-shockwave-flash") != 0)
    return;

  if (uri == NULL)
    return;

  if (video->media_url)
    g_free(video->media_url);

  video->media_url = g_strdup(uri);
}
#endif


/*
 * Make request on http to load the thumbnails
 */
void
video_load_thumbnails(Video *video, SoupSession *http_session)
{
  g_list_foreach(video->thumbnails, (GFunc)video_thumbnail_load_with_session, http_session);
}




/*
 * Take the next image in the list, load it if it's not into the cache, and
 * then display it. Move the next pointer to the next image into the list.
 *
 * We need to use a loop, because when we reach the end of the list, we need
 * to start again.
 *
 */
void
video_update_thumbnail(Video *video, SoupSession *session)
{
  GList *iter;
  gint loop;

  if (video->preview_image == NULL)
    return;

  if (video->thumbnail_is_loading == TRUE)
    return;

  for (loop=0; loop<2; loop++)
   {
     iter = video->next_thumbnail;
     while (iter)
      {
	VideoThumbnail *thumbnail = iter->data;

	if (thumbnail->pixbuf == NULL && thumbnail->error == 0)
	 {
	   /* Queue the image, when the image will be loaded, then this will be automatically display */
	   video_thumbnail_load_with_session(thumbnail, session);
	   video->thumbnail_is_loading = TRUE;
	   video->next_thumbnail = iter->next;
	   return;
	 }

	/* In case of an error, thumbnail->pixbuf is NULL */
	if (thumbnail->pixbuf)
	 {
	   trace("uri=%s width=%d height=%d\n", thumbnail->uri, thumbnail->width, thumbnail->height);
	   gtk_image_set_from_pixbuf(GTK_IMAGE(video->preview_image), thumbnail->pixbuf);
	   video->next_thumbnail = iter->next;
	   return;
	 }

	iter = iter->next;
      }

     /* We have reached the end of the list, restart from the beggining */
     video->next_thumbnail = video->thumbnails; 
   }

  /* No thumbnail can be used ... */
  video->next_thumbnail = video->thumbnails; 
}

static void
video_prepare_streaming_error(YoutubeVideo *yvideo, Video *video)
{
  error("Failed to get the stream\n");


}


static void
video_prepare_streaming_with_media_url(YoutubeVideo *yvideo, Video *video)
{
  char *uri;
  
  uri = youtube_video_get_stream_video_url(yvideo);
  if (uri == NULL)
   {
     trace("Failed to get video stream url\n");
     return;
   }
  g_signal_emit(video, video_signals[SIGNAL_VIDEO_READY], 0, uri);
}


/*
 * Try to get the media url for playing,
 * Return True if the media has been queue ...
 */
gboolean
video_prepare_streaming(Video *video, SoupSession *http_session)
{
  trace("webpage = %s\n", video->webpage);

  if (video->webpage == NULL)
    return FALSE;

  /* FIXME: Until with have a type for the website, use the url to detect the site */
  if (strstr(video->webpage, "youtube.com"))
   {
     YoutubeVideo *yvideo = youtube_video_new();

     if (youtube_video_set_video_url(yvideo, video->webpage) == FALSE)
       return FALSE;

     g_signal_connect(yvideo, "error", G_CALLBACK(video_prepare_streaming_error), video);
     g_signal_connect(yvideo, "stream-url-ready", G_CALLBACK(video_prepare_streaming_with_media_url), video);
     return TRUE;
   }
  else if (strncmp(video->webpage, "rtsp://", strlen("rtsp://")) == 0)
   {
     /* We try to setup a gstreamer pipe ... */
     g_signal_emit(video, video_signals[SIGNAL_VIDEO_READY], 0, video->webpage);
   }

  return FALSE;
}

void
video_set_state(Video *video, gboolean is_selected)
{
  static int has_saved_the_original_style_bg;

  if (has_saved_the_original_style_bg == 0)
   {
     GtkStyle *style = gtk_widget_get_style(video->evbox);
     memcpy(&color_normal, &style->bg[GTK_STATE_NORMAL], sizeof(color_normal));
     has_saved_the_original_style_bg = 1;
   }

  if (video->is_selected == is_selected)
    return;

  video->is_selected = is_selected;
  if (is_selected)
   {
     gtk_widget_modify_bg(video->evbox, GTK_STATE_NORMAL, &color_when_selected);
   }
  else
   {
     gtk_widget_modify_bg(video->evbox, GTK_STATE_NORMAL, &color_normal);
   }


}







/*
 *
 * .--------------------.
 * | XXXX_ 12450 views  | 
 * '--------------------'
 *
 */
static GtkWidget *
video_create_widget_score(Video *video)
{
  GtkWidget *hbox, *image, *label;
  unsigned int nstars;
  unsigned int i;
  char *text;

  hbox = gtk_hbox_new(FALSE, 0);

  nstars = rint(video->avg_rating);
  if (nstars > 5)
    nstars = 5;
  for (i=0; i<nstars; i++)
   {
     image = gtk_image_new();
     gtk_image_set_from_pixbuf(GTK_IMAGE(image), gold_star_pixbuf);
     gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 2);
   }
  for (;i<5; i++)
   {
     image = gtk_image_new();
     gtk_image_set_from_pixbuf(GTK_IMAGE(image), grey_star_pixbuf);
     gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 2);
   }

  text = g_strdup_printf("<span weight=\"light\">%u views</span>", video->views);
  label = gtk_label_new(NULL);
  gtk_label_set_markup(GTK_LABEL(label), text);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 8);
  g_free(text);

  if (video->num_votes)
   {
     text = g_strdup_printf(_("Ratings: %u\nAverage Rating: %.2lf"), video->num_votes, video->avg_rating);
     gtk_widget_set_tooltip_text(hbox, text);
   }
  else
   {
     text = g_strdup_printf(_("Average Rating: %.2lf"), video->avg_rating);
     gtk_widget_set_tooltip_text(hbox, text);
   }

  return hbox;
}

static GtkWidget *
video_create_widget_title(Video *video)
{
  char *markup;
  GtkWidget *hbox, *label;

  hbox = gtk_hbox_new(FALSE, 0);

  label = gtk_label_new(NULL);
  markup = g_markup_printf_escaped("<b>%s</b>", video->title);
  gtk_label_set_markup(GTK_LABEL(label), markup);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
  gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
  g_free(markup);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);

  return hbox;
}

static GtkWidget *
video_create_widget_info(Video *video)
{
  char *markup;
  GtkWidget *hbox, *label;
  char date_published[128];
  
  strftime(date_published, sizeof(date_published), "%c", localtime(&video->date_published));

  hbox = gtk_hbox_new(FALSE, 0);

  label = gtk_label_new(NULL);
  markup = g_markup_printf_escaped("<small>%s</small>\n<small>By <b>%s</b></small>",
     				   date_published,
				   video->author_name);
  gtk_label_set_markup(GTK_LABEL(label), markup);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
  g_free(markup);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);

  return hbox;


}


/*
 *  .---------------------------------.
 *  | Title of the video......        |
 *  |---------------------------------|
 *  |          | Wed, 09 Sep 2009     |
 *  |          | author               |
 *  |          | .........            |
 *  '---------------------------------'
 *
 *
 */
GtkWidget *
video_create_widget(Video *video)
{
  GtkWidget *hbox, *image, *vbox, *align, *evbox;
  

  /** */
  video->preview_image = image = gtk_image_new();
  gtk_image_set_from_pixbuf(GTK_IMAGE(image), black_pixbuf);


  /** */
  if (video->description)
   {
     align = gtk_alignment_new(0,0,0,0);
     gtk_container_add(GTK_CONTAINER(align), image);
     gtk_container_set_border_width(GTK_CONTAINER(align), 4);

     vbox = gtk_vbox_new(FALSE, 0);
     gtk_box_pack_start(GTK_BOX(vbox), video_create_widget_score(video), FALSE, FALSE, 4);
     gtk_box_pack_start(GTK_BOX(vbox), video_create_widget_info(video), FALSE, FALSE, 4);

     /** */
     hbox = gtk_hbox_new(FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 4);
     gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 4);

     vbox = gtk_vbox_new(FALSE, 0);
     gtk_box_pack_end(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
   }
  else
   {
     vbox = gtk_vbox_new(FALSE, 0);
     gtk_box_pack_end(GTK_BOX(vbox), image, TRUE, TRUE, 4);
   }


  /** */
  gtk_box_pack_start(GTK_BOX(vbox), video_create_widget_title(video), FALSE, FALSE, 4);

  video->evbox = evbox = gtk_event_box_new();

  gtk_container_add(GTK_CONTAINER(evbox), vbox);

  gtk_widget_show_all(evbox);

  return evbox;
}













Video *
video_new(void)
{
  Video *video = g_object_new(VIDEO_TYPE, NULL);

  return video;
}





/***********************************************************************
 *
 *
 *
 *
 * GObject things
 *
 *
 *
 *
 **********************************************************************/
G_DEFINE_TYPE(Video, video, G_TYPE_OBJECT)

static void
video_dispose(GObject *object)
{
  Video *video = (Video *)(object);

  trace("\n");

  if (video->preview_image)
   {
     gtk_widget_destroy(video->preview_image);
     video->preview_image = NULL;
   }

  if (video->webpage)
    g_free(video->webpage);

  if (video->author_name)
    g_free(video->author_name);

  if (video->title)
    g_free(video->title);

  if (video->description)
    g_free(video->description);


  g_list_foreach(video->thumbnails, (GFunc)g_object_unref, NULL);
  g_list_free(video->thumbnails);
  video->thumbnails = NULL;

  g_list_foreach(video->previews, (GFunc)g_object_unref, NULL);
  g_list_free(video->previews);
  video->previews = NULL;

  G_OBJECT_CLASS (video_parent_class)->dispose(object);
}

static void
video_class_init(VideoClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

  gobject_class->dispose = video_dispose;

  video_signals[SIGNAL_VIDEO_READY] =
        g_signal_new("video-ready",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1,
		     G_TYPE_STRING);


  /* set a black wallpaper if nothing else is available */
  black_pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
				FALSE,
				8,
				VIDEO_THUMBNAIL_WIDTH, VIDEO_THUMBNAIL_HEIGHT);
  gdk_pixbuf_fill(black_pixbuf, 255);

  /* load gold, grey star for the rating */
  gold_star_pixbuf = gtk_utils_get_pixbuf_from_icon_name("star.png", 16);
  grey_star_pixbuf = gdk_pixbuf_copy(gold_star_pixbuf);
  gdk_pixbuf_saturate_and_pixelate(grey_star_pixbuf, grey_star_pixbuf, 0.0, FALSE);

}

static void
video_init(Video *search)
{
  trace("\n");

}




