/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gst/gst.h>
#include <gst/interfaces/xoverlay.h>
#include <gst/video/video.h>



#include "gtktube.h"
#include "gtk_utils.h"
#include "video.h"
#include "video_view.h"
#include "utils.h"

enum
{
  SIGNAL_UNKNOWN_TYPE = 0,
  SIGNAL_BUFFERING,
  SIGNAL_END_OF_STREAM,
  SIGNAL_STATE_CHANGED,
  SIGNAL_LAST
};

static gint video_view_signals[SIGNAL_LAST];
static GdkGC *color_black_gc;





static gboolean
video_view_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
  VideoView *view = (VideoView *)(widget);

  if (view->pipeline == NULL)
   {
     trace("drawing black window\n");
     if (color_black_gc == NULL)
      {
	GdkColor color_black;
	gdk_color_parse("black", &color_black);
	color_black_gc = gdk_gc_new(GDK_DRAWABLE(widget->window));
	gdk_gc_set_rgb_fg_color(color_black_gc, &color_black);
      }
     gdk_draw_rectangle(widget->window, color_black_gc, TRUE,
			0, 0, widget->allocation.width, widget->allocation.height);

   }
  else if (opt_no_video == 0)
   {
     trace("gst_x_overlay_set_xwindow_id()\n");
     gst_x_overlay_set_xwindow_id(GST_X_OVERLAY(view->videosink), GDK_WINDOW_XWINDOW(widget->window));
   }

  return FALSE;
}

#if 0
static void
video_view_on_unknown_type_cb(GstElement              *decodebin,
			      GstPad                  *pad,
			      GstCaps                 *caps,
			      gpointer                 user_data)
{
  VideoView *view = user_data;

  trace("\n");
  g_signal_emit(view, video_view_signals[SIGNAL_UNKNOWN_TYPE], 0); 
}
#endif

static void
video_view_get_caps(VideoView *view)
{
  GList   *streaminfo;
  GstPad  *pad;

  g_return_if_fail(view);
  g_return_if_fail(view->playbin);

  g_object_get(view->playbin, "stream-info", &streaminfo, NULL);

  while (streaminfo)
   {
     gint type;
     GParamSpec *pspec;
     GEnumValue *val;
     GstStructure *s = NULL;
     GObject *info = streaminfo->data;

     g_object_get(info, "type", &type, NULL);
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(info), "type");
     val = g_enum_get_value(G_PARAM_SPEC_ENUM(pspec)->enum_class, type);

     g_object_get(info, "object", &pad, NULL);
     if (pad && GST_IS_PAD (pad) && GST_PAD_CAPS (pad))
      {
	s = gst_caps_get_structure(GST_PAD_CAPS (pad), 0);

	if (strcmp(val->value_nick, "audio") == 0)
	 {
	   view->has_audio = TRUE;
	   gst_structure_get_int(s, "channels", &view->audio_nchannels);
	   gst_structure_get_int(s, "rate", &view->audio_rate);

	 }
	else if (strcmp (val->value_nick, "video") == 0)
	 {
	   view->has_video = TRUE;
	   gst_structure_get_int(s, "width", &view->video_width);
	   gst_structure_get_int(s, "height", &view->video_height);
	   gst_structure_get_double(s, "framerate", &view->video_fps);
	 }
      }

     streaminfo = streaminfo->next;
   }

  g_list_free(streaminfo);

  trace("Video: %dx%d @%.0lffps  Audio: %d channels at %.1fkHz\n",
        view->video_width, view->video_height, view->video_fps,
	view->audio_nchannels, (float)view->audio_rate/1000);

  if (view->has_video && opt_no_video == 0)
   {
     if (   (view->video_height != view->height && view->video_height > 0)
	 || (view->video_width != view->width && view->video_width > 0))
      {
	gtk_widget_set_size_request(GTK_WIDGET(view), view->video_width, view->video_height);
      }
   }

}




static gboolean
video_view_gstreamer_bus_callback(GstBus *bus, GstMessage *message, gpointer user_data)
{
  VideoView *view = user_data;

  switch (GST_MESSAGE_TYPE (message))
   {
    case GST_MESSAGE_ERROR:
       {
	 GError *err;
	 gchar *debug;

	 gst_message_parse_error (message, &err, &debug);
	 g_print ("Error: %s\n", err->message);
	 g_error_free (err);
	 g_free(debug);

	 break;
       }

    case GST_MESSAGE_BUFFERING:
       {
	 int percent = 0;
	 gst_message_parse_buffering(message, &percent);
	 trace("Buffering: %d%%\n", percent);

	 g_signal_emit(view, video_view_signals[SIGNAL_BUFFERING], 0, percent); 
	 break;
       }

    case GST_MESSAGE_STATE_CHANGED:
       {
	 GstState old, new;

	 gst_message_parse_state_changed(message, &old, &new, NULL);
	 if (old == new)
	   break;

	 /* we only care about playbin (pipeline) state changes */
	 if (GST_MESSAGE_SRC(message) != GST_OBJECT(view->playbin))
	   break;

	 if (new == GST_STATE_READY)
	   trace("GST_STATE_READY\n");
	 else if (new == GST_STATE_PAUSED)
	   trace("GST_STATE_PAUSED\n");
	 else if (new == GST_STATE_PLAYING)
	   trace("GST_STATE_PLAYING\n");

	 /* Update caps */
	 if (old == GST_STATE_READY && new == GST_STATE_PAUSED)
	   video_view_get_caps(view);

	 g_signal_emit(view, video_view_signals[SIGNAL_STATE_CHANGED], 0, (gint)new); 
	 break;
       }

    case GST_MESSAGE_EOS:
      /* end-of-stream */
      trace("END OF STREAM\n");
      g_signal_emit(view, video_view_signals[SIGNAL_END_OF_STREAM], 0); 
      break;

#if 0
    case GST_MESSAGE_TAG:
       {
	 GstTagList *tag_list, *result;
	 GstElementFactory *f;

	 gst_message_parse_tag (message, &tag_list);

	 trace("GST_MESSAGE_TAG\n");

	 trace ("Tags: %" GST_PTR_FORMAT " \n", tag_list);


	 break;
       }
#endif

    default:
      /* unhandled message */
      trace("unhandled message: %s\n", GST_MESSAGE_TYPE_NAME(message));
      break;
   }

  /* remove message from the queue */
  return TRUE;
}






static gboolean
video_view_create_pipeline(VideoView *view, const char *media_url)
{
  GstElement *pipeline;
  GstElement *src = NULL;
  GstElement *videosink;
  GstBus *bus;

  view->has_audio = FALSE;
  view->has_video = FALSE;
  view->video_width = 0;
  view->video_height = 0;

  pipeline = gst_pipeline_new("pipe");

  if (opt_no_video)
    videosink = gst_element_factory_make("fakesink", NULL);
  else
   {
     videosink = gst_element_factory_make("xvimagesink", NULL);
     g_object_set(G_OBJECT(videosink), "sync", TRUE, NULL);
     g_object_set(G_OBJECT(videosink), "force-aspect-ratio", TRUE, NULL);
   }

  src = gst_element_factory_make("playbin", NULL);
  g_object_set(G_OBJECT(src),
     	       "uri", media_url,
	       "video-sink", videosink,
	       NULL);
  view->playbin = src;

  gst_bin_add_many(GST_BIN(pipeline), src, NULL);

  bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
  gst_bus_add_watch(bus, video_view_gstreamer_bus_callback, view);
  gst_object_unref(bus);

  view->pipeline = pipeline;

  if (opt_no_video == 0 && GTK_WIDGET_DRAWABLE(GTK_WIDGET(view)))
   {
     gst_x_overlay_set_xwindow_id(GST_X_OVERLAY(videosink),
				  GDK_WINDOW_XWINDOW(GTK_WIDGET(view)->window));
   }

  view->pipeline = pipeline;
  view->videosink = videosink;

  return TRUE;

}


static void
video_view_on_video_ready(Video *video, const char *media_url, VideoView *view)
{
  trace("media_url = %s\n", media_url);

  if (view->video)
    g_object_unref(view->video);

  view->video = g_object_ref(video);

  /* Setup the stream */
  //video_view_setup_http_streamer(view, NULL, media_url);
  video_view_create_pipeline(view, media_url);

  /* Play the video by default */
  video_view_play(view);
}




/*
 * Some video need to make request before being to play.
 * So when the video is ready a signal is emited to tell that the video is ready.
 *
 *
 */
void
video_view_set_video(VideoView *view, Video *video, SoupSession *http_session)
{
  trace("webpage = %s\n", video->webpage);
  g_signal_connect(video, "video-ready", G_CALLBACK(video_view_on_video_ready), view);
  video_prepare_streaming(video, http_session);
}

void
video_view_play(VideoView *view)
{
  GstStateChangeReturn ret;

  g_return_if_fail(view->pipeline != NULL);

  ret = gst_element_set_state(view->pipeline, GST_STATE_PLAYING);

  if (ret == GST_STATE_CHANGE_FAILURE)
   {
     g_print ("Failed to start up pipeline!\n");
   }

}

void
video_view_stop(VideoView *view)
{
  if (view->pipeline == NULL)
    return;

  gst_element_set_state(view->pipeline, GST_STATE_NULL);

  g_object_unref(view->pipeline);
  view->pipeline = NULL;
  view->playbin = NULL;
  view->videosink = NULL;
}

void
video_view_pause(VideoView *view)
{
  g_return_if_fail(view->pipeline != NULL);

  gst_element_set_state(view->pipeline, GST_STATE_PAUSED);
}



void
video_view_get_times(VideoView *view, gint64 *position, gint64 *duration)
{
  GstFormat fmt = GST_FORMAT_TIME;

  if (view->pipeline == NULL)
   {
     *position = -1;
     *duration = -1;
     return;
   }

  if (gst_element_query_position(view->pipeline, &fmt, position) == FALSE)
    *position = -1;

  if (gst_element_query_duration(view->pipeline, &fmt, duration) == FALSE)
    *duration = -1;
}

/*
 *
 *
 *
 */
GtkWidget *
video_view_new(void)
{
  VideoView *view = g_object_new(VIDEO_VIEW_TYPE, NULL);

  return (GtkWidget *)view;
}






/***********************************************************************
 *
 *
 *
 *
 * GObject things
 *
 *
 *
 *
 **********************************************************************/
G_DEFINE_TYPE(VideoView, video_view, GTK_TYPE_DRAWING_AREA)

static void
video_view_dispose(GObject *object)
{
  VideoView *view = (VideoView *)(object);

  trace("\n");

  if (view->video)
    g_object_unref(view->video);
  view->video = NULL;


  G_OBJECT_CLASS (video_view_parent_class)->dispose(object);
}

static void
video_view_class_init(VideoViewClass *klass)
{
  GObjectClass *gobject_class;
  GtkWidgetClass *widget_class;

  gobject_class   = (GObjectClass *)klass;
  widget_class    = (GtkWidgetClass *)klass;


  gobject_class->dispose = video_view_dispose;

 // widget_class->expose_event = video_view_expose;

  video_view_signals[SIGNAL_UNKNOWN_TYPE] =
        g_signal_new("unknown-type",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

  video_view_signals[SIGNAL_END_OF_STREAM] =
        g_signal_new("end-of-stream",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

  video_view_signals[SIGNAL_BUFFERING] =
        g_signal_new("buffering",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT,
                     G_TYPE_NONE, 1,
		     G_TYPE_INT);

  video_view_signals[SIGNAL_STATE_CHANGED] =
        g_signal_new("video-state-changed",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__INT,
                     G_TYPE_NONE, 1,
		     G_TYPE_INT);

}

static void
video_view_init(VideoView *view)
{
  trace("\n");

  view->width = VIDEO_VIEW_WIDTH;
  view->height = VIDEO_VIEW_HEIGHT;

  gtk_widget_set_size_request(GTK_WIDGET(view), view->width, view->height);
  g_signal_connect(GTK_OBJECT(view), "expose-event", G_CALLBACK(video_view_expose), NULL);
}




