/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gtktube.h"
#include "gtktube_search.h"
#include "youtube_search.h"
#include "gtk_utils.h"
#include "utils.h"
#include "marshal.h"


#define DEBUG_XML 0



enum
{
  SIGNAL_SEARCH_PROGRESS_STATUS,
  SIGNAL_SEARCH_FINISHED,
  SIGNAL_LAST
};

static gint youtube_search_signals[SIGNAL_LAST];



/*
 *
 * XML functions to parse result from youtube
 *
 *
 */


/* Called for open tags <foo bar="baz"> */
static void
youtube_xml_parser_start_element(GMarkupParseContext  *context,
				 const gchar          *element_name,
				 const gchar         **attribute_names,
				 const gchar         **attribute_values,
				 gpointer              user_data,
				 GError              **error)
{
  YoutubeSearch *search = user_data;
  unsigned int i;

#if DEBUG_XML
  static char indent_space[17] = "                \0";
  g_debug("%s<%s>", indent_space+16-search->xml_indent, element_name);
  search->xml_indent++;
#endif

  if (search->xml_state == YOUTUBE_XML_STATE_UNKNOWN && strcasecmp(element_name, "FEED") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_FEED;
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_FEED && strcasecmp(element_name, "ENTRY") == 0)
   {
     search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY;

     g_return_if_fail(search->current_video == NULL);

     search->current_video = video_new();
   }
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY)
   {
     /* 
      * <feed>
      *   <entry>
      *     <link>
      *     <link rel='self' type='application/atom+xml' href='http://gdata.youtube.com/feeds/api/videos/lPT_3PEjnsE'/>
      *     <link rel='http://gdata.youtube.com/schemas/2007#mobile' type='text/html' href='http://m.youtube.com/details?v=lPT_3PEjnsE'/>
      *     <link rel='alternate' type='text/html' href='http://www.youtube.com/watch?v=lPT_3PEjnsE'/>
      *
      * Link tag need 'rel' and 'type' attributes
      */
     if (strcmp(element_name, "link") == 0)
      {
	unsigned int matches = 0;
	const char *href = NULL;
	for (i=0; attribute_names[i]; i++)
	 {
	   if ((strcmp(attribute_names[i], "rel") == 0) && (strcmp(attribute_values[i], "alternate") == 0))
	     matches++;
	   else if ((strcmp(attribute_names[i], "type") == 0) && (strcmp(attribute_values[i], "text/html") == 0))
	     matches++;
	   else if (strcmp(attribute_names[i], "href") == 0)
	    {
	      matches++;
	      href = attribute_values[i];
	    }
	 }
	if (matches == 3)
	  video_set_webpage(search->current_video, href);
      }

     /* 
      * <feed>
      *   <entry>
      *     <author>
      *       ...
      *
      */
     else if (strcmp(element_name, "author") == 0)
      {
	search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR;
      }

     /* 
      * <feed>
      *   <entry>
      *     <published>
      *       ISO-DATE
      *     </published>
      *
      */
     else if (strcmp(element_name, "published") == 0)
      {
	search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_PUBLISHED;
      }

     /* 
      * <feed>
      *   <entry>
      *     <yt:statistics favoriteCount='65524' viewCount='10284820'/>
      *
      */
     else if (strcmp(element_name, "yt:statistics") == 0)
      {
      	unsigned int views = 0;
      	unsigned int favorites = 0;
	for (i=0; attribute_names[i]; i++)
	 {
	   if (strcmp(attribute_names[i], "favoriteCount") == 0)
	     sscanf(attribute_values[i], "%u", &favorites);
	   else if (strcmp(attribute_names[i], "viewCount") == 0)
	     sscanf(attribute_values[i], "%u", &views);
	 }
	video_set_statistics(search->current_video, views, favorites);
      }

     /* 
      * <feed>
      *   <entry>
      *     <gd:rating average='4.4545455' max='5' min='1' numRaters='22' rel='[...]'>
      *
      */
     else if (strcmp(element_name, "gd:rating") == 0)
      {
      	gdouble average = 0;
	int max_score = 0;
	int min_score = 0;
	int num_votes = 0;
	for (i=0; attribute_names[i]; i++)
	 {
	   if (strcmp(attribute_names[i], "average") == 0)
	     average = g_ascii_strtod(attribute_values[i], NULL);
	   else if (strcmp(attribute_names[i], "max") == 0)
	     sscanf(attribute_values[i], "%u", &max_score);
	   else if (strcmp(attribute_names[i], "min") == 0)
	     sscanf(attribute_values[i], "%u", &min_score);
	   else if (strcmp(attribute_names[i], "numRaters") == 0)
	     sscanf(attribute_values[i], "%u", &num_votes);
	 }
	video_set_ratings(search->current_video, average, min_score, max_score, num_votes);
      }


     /* 
      * <feed>
      *   <entry>
      *     <media:group>
      *       ...
      *
      */
     else if (strcmp(element_name, "media:group") == 0)
      {
	search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP;
      }


   }
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR)
   {
     /* 
      * <feed>
      *   <entry>
      *     <author>
      *       <name>
      *
      */
     if (strcmp(element_name, "name") == 0)
      {
	/* Need to get the content of name ... */
	search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR_NAME;
      }

   }

  /* 
   * <feed>
   *   <entry>
   *     <media:group>
   *       <media:category label='Music' scheme='http://gdata.youtube.com/schemas/2007/categories.cat'>Music</media:category>
   *       <media:content url='http://www.youtube.com/v/lPT_3PEjnsE' type='application/x-shockwave-flash' medium='video' [...]
   *       <media:content url='rtsp://rtsp2.youtube.com/[...]/video.3gp' type='video/3gpp' medium='video'  [...]
   *       <media:description type='plain'>[...]
   *       <media:thumbnail url='http://i.ytimg.com/vi/lPT_3PEjnsE/2.jpg' height='90' width='120' time='00:02:16'/>
   *       <media:title type='plain'>Toto - "Africa" Music Video</media:title>
   *       <yt:duration seconds='272'/>
   *       ...
   *
   */
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP)
   {
     /* 
      * <feed>
      *   <entry>
      *     <media:group>
      *       <media:thumbnail url='http://i.ytimg.com/vi/lPT_3PEjnsE/2.jpg' height='90' width='120' time='00:02:16'/>
      */
     if (strcmp(element_name, "media:thumbnail") == 0)
      {
    	const char *uri = NULL;
	int width = 0;
	int height = 0;
	for (i=0; attribute_names[i]; i++)
	 {
	   if (strcmp(attribute_names[i], "width") == 0)
	     sscanf(attribute_values[i], "%u", &width);
	   else if (strcmp(attribute_names[i], "height") == 0)
	     sscanf(attribute_values[i], "%u", &height);
	   else if (strcmp(attribute_names[i], "url") == 0)
	     uri = attribute_values[i];
	 }
	
	video_add_thumbnail_uri(search->current_video, uri, width, height);
      }

     /* 
      * <feed>
      *   <entry>
      *     <media:group>
      *       <media:title type='plain'>Toto - "Africa" Music Video</media:title>
      */
     else if (strcmp(element_name, "media:title") == 0)
      {
	search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP_TITLE;
      }

     /* 
      * <feed>
      *   <entry>
      *     <media:group>
      *       <media:description type='plain'>[...]
      */
     else if (strcmp(element_name, "media:description") == 0)
      {
	search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP_DESCRIPTION;
      }

     /* 
      * <feed>
      *   <entry>
      *     <media:group>
      *       <yt:duration seconds='272'/>
      */
     else if (strcmp(element_name, "yt:duration") == 0)
      {
      	int duration = 0;
	for (i=0; attribute_names[i]; i++)
	 {
	   if (strcmp(attribute_names[i], "seconds") == 0)
	     sscanf(attribute_values[i], "%u", &duration);
	 }
	
	video_set_duration(search->current_video, duration);
      }

     /* 
      * <feed>
      *   <entry>
      *     <media:group>
      *       <media:content url='http://www.youtube.com/v/lPT_3PEjnsE' type='application/x-shockwave-flash' medium='video' [...]
      *       <media:content url='rtsp://rtsp2.youtube.com/[...]/video.3gp' type='video/3gpp' medium='video'  [...]
      */
     else if (strcmp(element_name, "media:content") == 0)
      {
    	const char *uri = NULL;
    	const char *type = NULL;
	guint format = 0;

	for (i=0; attribute_names[i]; i++)
	 {
	   if (strcmp(attribute_names[i], "type") == 0)
	     type = attribute_values[i];
	   else if (strcmp(attribute_names[i], "yt:format") == 0)
	     sscanf(attribute_values[i], "%u", &format);
	   else if (strcmp(attribute_names[i], "url") == 0)
	     uri = attribute_values[i];
	 }
	
#if 0
	video_add_content(search->current_video, uri, type, format);
#endif
      }

   }

}

/* Called for close tags </foo> */
static void
youtube_xml_parser_end_element(GMarkupParseContext *context,
			       const gchar         *element_name,
			       gpointer             user_data,
			       GError             **error)
{
  YoutubeSearch *search = user_data;

#if DEBUG_XML
  static char indent_space[17] = "                \0";
  search->xml_indent--;
  g_debug("%s</%s>  (%d)", indent_space+16-search->xml_indent, element_name, search->xml_state);
#endif

  if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY && strcmp(element_name, "entry") == 0)
   {
     if (search->current_video)
      {
	search->videos = g_list_prepend(search->videos, search->current_video);
	search->current_video = NULL;
      }
     else
      {
	g_error("search->current_video == NULL\n");
      }

     search->xml_state = YOUTUBE_XML_STATE_IN_FEED;
   }

  if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR_NAME && strcmp(element_name, "name") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR;
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR && strcmp(element_name, "author") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY;
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_PUBLISHED && strcmp(element_name, "published") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY;
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP && strcmp(element_name, "media:group") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY;

  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP_TITLE && strcmp(element_name, "media:title") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP;
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP_DESCRIPTION && strcmp(element_name, "media:description") == 0)
    search->xml_state = YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP;


}

/* Called for character data (/!\ text is not nul-terminated) */
static void
youtube_xml_parser_text_handler(GMarkupParseContext *context,
				const gchar         *text,
				gsize                text_len,  
				gpointer             user_data,
				GError             **gerror)
{
  YoutubeSearch *search = user_data;

  /* Speedup the processing */
  if (text_len == 0)
    return;

  if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY)
   {
#if DEBUG_XML
      static char indent_space[17] = "                \0";
      char *content = g_strndup(text, text_len>30?30:text_len);
      g_debug("%s[%s]", indent_space+16-search->xml_indent, content);
      g_free(content);
#endif
   }
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_AUTHOR_NAME)
   {
     video_set_author_name(search->current_video, text, text_len);
   }
  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_PUBLISHED)
   {
     SoupDate *d;
     /* Add the '\0' */
     char date_string[128];
     if (text_len > 127)
       text_len = 127;
     memcpy(date_string, text, text_len);
     date_string[text_len] = 0;

     d = soup_date_new_from_string(date_string);
     if (d == NULL)
      {
	error("Date can't be parsed (%s)\n", date_string);
	return;
      }

     video_set_published_date(search->current_video, soup_date_to_time_t(d));
     soup_date_free(d);
   }

  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP_TITLE)
   {
     video_set_title(search->current_video, text, text_len);
   }

  else if (search->xml_state == YOUTUBE_XML_STATE_IN_ENTRY_MEDIA_GROUP_DESCRIPTION)
   {
     video_set_description(search->current_video, text, text_len);
   }

}


static void
youtube_xml_parser_error(GMarkupParseContext *context,
			 GError              *error,
			 gpointer             user_data)
{
  //YoutubeSearch *search = user_data;

  /* TODO: copy GError */
  g_warning("Error in parsing XML 13833\n");
}

static GMarkupParser youtube_xml_parser =
{
  youtube_xml_parser_start_element,
  youtube_xml_parser_end_element,
  youtube_xml_parser_text_handler,
  NULL,
  youtube_xml_parser_error
};


static void
youtube_search_parse_xml_response(YoutubeSearch *search, SoupMessage *msg)
{
  GError *gerror = NULL;
  GMarkupParseContext *context;
  const char *data = msg->response_body->data;
  gint datalen = msg->response_body->length;

  search->xml_state = YOUTUBE_XML_STATE_UNKNOWN;

  context = g_markup_parse_context_new(&youtube_xml_parser, 0, search, NULL);

  g_markup_parse_context_parse(context, data, datalen, &gerror);
  if (gerror)
   {
     error("Error while parsing response from Youtube. (%s)\n", gerror->message);
     g_markup_parse_context_free(context);
     g_error_free(gerror);
     return;
   }

  if (g_markup_parse_context_end_parse(context, &gerror) == FALSE)
   {
     error("Error while parsing response from Youtube at the end. (%s)\n", gerror->message);
     g_markup_parse_context_free (context);
     g_error_free(gerror);
     return;
   }

  g_markup_parse_context_free (context);
  
}




/*
 *
 * Networks layers to make request
 *
 *
 *
 */

static void
youtube_search_async_callback(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
  YoutubeSearch *search = user_data;

  trace("\n");

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code))
   {
     /* Youtube return error in the body */
     search->http_return_code = msg->status_code;
   } 
  else
   {
     search->http_return_code = 0;
   }

  trace("download status code: %d\n", msg->status_code);
  trace("%d bytes downloaded\n", (gsize)msg->response_body->length);
  trace("%s\n--\n", msg->response_body->data);

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code) == FALSE)
   {
     youtube_search_parse_xml_response(search, msg);
   }

  gdk_threads_enter();
  g_signal_emit(search, youtube_search_signals[SIGNAL_SEARCH_FINISHED], 0);
  gdk_threads_leave();
}

static void
youtube_search_got_chunk_cb(SoupMessage *msg, SoupBuffer *chunk, gpointer user_data)
{
  YoutubeSearch *search = user_data;
  guint64 content_length;
  guint64 received_content_length;

  if (msg->response_body == NULL)
    return;

  content_length = soup_message_headers_get_content_length(msg->response_headers);
  received_content_length = msg->response_body->length;
  trace("got = %lld/%lld\n", received_content_length, content_length);

  gdk_threads_enter();
  g_signal_emit(search, youtube_search_signals[SIGNAL_SEARCH_PROGRESS_STATUS], 0, content_length, received_content_length);
  gdk_threads_leave();
}



static void
youtube_search_prepare_search(YoutubeSearch *search, const char *keywords)
{
  GString *url;
  SoupMessage *msg;
  char *q_keywords;

  trace("\n");

  if (search->msg)
   {
     g_warning("Clearing the previous Youtube query ...\n");
     g_object_unref(search->msg);
     search->msg = NULL;
   }

  url = g_string_new(NULL);
  q_keywords = soup_uri_encode(keywords, NULL);
  g_string_printf(url, "http://gdata.youtube.com/feeds/api/videos?q=%s&max-results=%d&start-index=%d",
      		  q_keywords,
		  search->max_result,
		  search->start_index);
  g_free(q_keywords);

  switch (search->search_order)
   {
    case SORT_BY_NEWEST:
      g_string_append(url, "&orderby=published");
      break;

    default:
    case SORT_BY_COUNT:
    case SORT_BY_POPULARITY:
      g_string_append(url, "&orderby=viewCount");
      break;
   }


  msg = soup_message_new("GET", url->str);
  g_string_free(url, TRUE);
  if (msg == NULL)
    return;

  g_signal_connect(G_OBJECT(msg), "got-chunk", G_CALLBACK(youtube_search_got_chunk_cb), search);


  if (search->async_http_session == NULL)
    search->async_http_session = soup_session_async_new();

  soup_session_queue_message(search->async_http_session, msg, youtube_search_async_callback, search);
}






/*
 *
 * Public interface
 *
 *
 */




YoutubeSearch *
youtube_search_new(void)
{
  YoutubeSearch *search = g_object_new(YOUTUBE_SEARCH_TYPE, NULL);

  return search;
}

gboolean
youtube_search_search_with_keywords(YoutubeSearch *search, const char *keywords)
{
  search->start_index = 1;

  youtube_search_prepare_search(search, keywords);

  return TRUE;
}

void
youtube_search_set_http_session(YoutubeSearch *search, SoupSession *http_session)
{
  search->async_http_session = g_object_ref(http_session);
}

GList *
youtube_search_get_results(YoutubeSearch *search)
{
  GList *videos = search->videos;

  return videos;
}


/***********************************************************************
 *
 *
 *
 *
 * GObject things
 *
 *
 *
 *
 **********************************************************************/
G_DEFINE_TYPE(YoutubeSearch, youtube_search, G_TYPE_OBJECT)

static void
youtube_search_finalize(GObject *object)
{
  YoutubeSearch *search = (YoutubeSearch *)(object);

  if (search->async_http_session)
    g_object_unref(search->async_http_session);
  search->async_http_session = NULL;

  if (search->msg)
    g_object_unref(search->msg);

  G_OBJECT_CLASS (youtube_search_parent_class)->finalize(object);
}

static void
youtube_search_class_init(YoutubeSearchClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

  gobject_class->finalize = youtube_search_finalize;

  youtube_search_signals[SIGNAL_SEARCH_FINISHED] =
        g_signal_new("search-finished",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

  youtube_search_signals[SIGNAL_SEARCH_PROGRESS_STATUS] =
        g_signal_new("progress-status",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     gtktube_marshal_VOID__UINT64_UINT64,
                     G_TYPE_NONE,
		     2,
		     G_TYPE_UINT64,
		     G_TYPE_UINT64);

}

static void
youtube_search_init(YoutubeSearch *search)
{
  trace("\n");

  search->async_http_session = NULL;
  search->max_result = 10;
  search->search_order = SORT_BY_COUNT;
}




/***********************************************************************
 *
 *
 *
 *
 * Plugin interface
 *
 *
 *
 *
 **********************************************************************/

static void
on_youtube_search_finished(YoutubeSearch *search, gpointer user_data)
{
  struct gtktube_search_plugin *priv = user_data;

  if (priv->search_finished_cb)
   {
     GList *videos = youtube_search_get_results(search);
     priv->search_finished_cb(videos, priv->search_finished_cb_priv_data);
   }

}

static void
on_youtube_progress_status(YoutubeSearch *search, guint64 content_length, guint64 received_content_length, gpointer user_data)
{
  struct gtktube_search_plugin *priv = user_data;

  if (priv->progress_status_cb)
    priv->progress_status_cb(content_length, received_content_length, priv->progress_status_cb_priv_data);

}


#define YOUTUBE_PLUGIN_NAME "youtube"

struct gtktube_search_plugin plugin_youtube =
{
   .name           = YOUTUBE_PLUGIN_NAME,
   .description    = "Youtube Videos",

   .type           = GTKTUBE_PLUGIN_SEARCH_ENGINE,
   .logo           = "youtube.png",

   .have_search    = TRUE,

   .set_http_session = (SearchPluginSetHttpSessionCallback)youtube_search_set_http_session,
   .search_for = (SearchPluginSearchForCallback)youtube_search_search_with_keywords,
};


static int
youtube_plugin_init(void)
{
  trace("\n");

  YoutubeSearch *priv  = youtube_search_new();
  plugin_youtube.priv = priv;

  g_signal_connect(priv, "search-finished", G_CALLBACK(on_youtube_search_finished), &plugin_youtube);
  g_signal_connect(priv, "progress-status", G_CALLBACK(on_youtube_progress_status), &plugin_youtube);

  return gtktube_search_plugin_register(&plugin_youtube);
}

static void
youtube_plugin_exit(void)
{
  trace("\n");
  gtktube_search_plugin_unregister(&plugin_youtube);
}



GTKTUBE_PLUGIN_DEFINE(youtube,
    		      "Youtube Search Engine",
		      "Luc Saillard <luc@saillard.org>",
		      PACKAGE_VERSION,
		      youtube_plugin_init,
		      youtube_plugin_exit)


