/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#if HAVE_JSONGLIB
#include <json-glib/json-glib.h>
#endif

#include "gtktube.h"
#include "youtube_suggest.h"
#include "gtk_utils.h"
#include "utils.h"
#include "marshal.h"




#if HAVE_JSONGLIB

enum
{
  SIGNAL_SUGGEST_FINISHED,
  SIGNAL_LAST
};

static gint youtube_suggest_signals[SIGNAL_LAST];

static const char *youtube_suggest_search_country_code = "en";



static GList *
youtube_suggest_parse_json_response(YoutubeSuggest *search, SoupMessage *msg)
{
  JsonParser *parser;
  JsonNode *root;
  GError *gerror = NULL;
  GList *autosuggest_list = NULL;
  JsonArray *jarray, *jarray_strings, *jarray_results;
  unsigned int i, n;


  parser = json_parser_new();
  json_parser_load_from_data(parser, msg->response_body->data, msg->response_body->length, &gerror);
  if (gerror)
   {
     trace("Unable to parse json data: %s\n", gerror->message);
     g_error_free(gerror);
     g_object_unref(parser);
     return autosuggest_list;
   }

  root = json_parser_get_root(parser);

  /*
   * Response:
   *   ["azaaaaaaaaaaaaaaaaaaaaaaaaaaaaaeeeeeeeeeee",[],[]]
   *   ["toto",["toto africa","toto cutugno","toto hold the line","toto caca","toto cutugno l'italiano","toto qui presente la meteo","totoro","toto rosanna","toto live","toto story"],["2 970 résultats","1 480 résultats","1 180 résultats","379 résultats","228 résultats","18 résultats","2 630 résultats","3 640 résultats","2 780 résultats","289 résultats"]]
   *   ["toto",["toto africa","toto hold the line","toto cutugno","toto rosanna","totoro","totoy bato","toto cotugno","toto live","toto cotugno l'italiano","toto i'll be over you"],["1,850 results","1,180 results","1,290 results","3,620 results","2,610 results","826 results","205 results","2,670 results","53 results","311 results"]]
   *
   */
  if (JSON_NODE_TYPE(root) != JSON_NODE_ARRAY)
   {
     trace("root node is not an array\n");
     g_object_unref(parser);
     return autosuggest_list;
   }

  jarray = json_node_get_array(root);
  if (jarray == NULL)
   {
     trace("root node is not an array\n");
     g_object_unref(parser);
     return autosuggest_list;
   }

  if (json_array_get_length(jarray) != 3)
   {
     trace("root node is not an array with 3 elements\n");
     g_object_unref(parser);
     return autosuggest_list;
   }

  jarray_strings = json_node_get_array(json_array_get_element(jarray, 1));
  jarray_results = json_node_get_array(json_array_get_element(jarray, 2));
  if (   (jarray_results == NULL)
      || (jarray_results == NULL)
      || (json_array_get_length(jarray_strings) != json_array_get_length(jarray_results)))
   {
     trace("Bad JSON response: not an array\n");
     g_object_unref(parser);
     return autosuggest_list;
   }

  n = json_array_get_length(jarray_strings);
  for (i=0; i<n; i++)
   {
#if 0
     char *text = g_strdup_printf("%s (%s)",
	                          json_array_get_string_element(jarray_strings, i),
	                          json_array_get_string_element(jarray_results, i));
#else
     char *text = g_strdup_printf("%s", json_array_get_string_element(jarray_strings, i));
#endif
     autosuggest_list = g_list_prepend(autosuggest_list, text);
   }
  autosuggest_list = g_list_reverse(autosuggest_list);

  g_object_unref(parser);
  return autosuggest_list;
}




/*
 *
 * Networks layers to make request
 *
 *
 *
 */

static void
youtube_suggest_async_callback(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
  YoutubeSuggest *search = user_data;
  GList *autosuggest_list;

  trace("\n");

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code))
   {
     /* Youtube return error in the body */
     search->http_return_code = msg->status_code;
   } 
  else
   {
     search->http_return_code = 0;
   }

  trace("download status code: %d\n", msg->status_code);
  trace("%d bytes downloaded\n", (gsize)msg->response_body->length);
  trace("%s\n--\n", msg->response_body->data);

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code) == FALSE)
   {
     autosuggest_list = youtube_suggest_parse_json_response(search, msg);
     search->autosuggest_list = autosuggest_list;
   }

  gdk_threads_enter();
  g_signal_emit(search, youtube_suggest_signals[SIGNAL_SUGGEST_FINISHED], 0);
  gdk_threads_leave();
}


static void
youtube_suggest_prepare_search(YoutubeSuggest *search, const char *keywords)
{
  GString *url;
  SoupMessage *msg;
  char *q_keywords;

  trace("\n");

  url = g_string_new(NULL);
  q_keywords = soup_uri_encode(keywords, NULL);
  g_string_printf(url, "http://suggestqueries.google.com/complete/search?hl=%s&ds=yt&json=t&q=%s",
      		  youtube_suggest_search_country_code,
      		  q_keywords);
  g_free(q_keywords);

  msg = soup_message_new("GET", url->str);
  g_string_free(url, TRUE);
  if (msg == NULL)
    return;

  if (search->async_http_session == NULL)
    search->async_http_session = soup_session_async_new();

  soup_session_queue_message(search->async_http_session, msg, youtube_suggest_async_callback, search);
}







/*
 *
 * Public interface
 *
 *
 */




YoutubeSuggest *
youtube_suggest_new(void)
{
  YoutubeSuggest *search = g_object_new(YOUTUBE_SUGGEST_TYPE, NULL);

  return search;
}

void
youtube_suggest_search_with_keywords(YoutubeSuggest *search, const char *keywords)
{
  youtube_suggest_prepare_search(search, keywords);

}

void
youtube_suggest_set_http_session(YoutubeSuggest *search, SoupSession *http_session)
{
  search->async_http_session = g_object_ref(http_session);
}

GList *
youtube_suggest_get_results(YoutubeSuggest *search)
{
  return search->autosuggest_list;
}


/***********************************************************************
 *
 *
 *
 *
 * GObject things
 *
 *
 *
 *
 **********************************************************************/
G_DEFINE_TYPE(YoutubeSuggest, youtube_suggest, G_TYPE_OBJECT)

static void
youtube_suggest_finalize(GObject *object)
{
  YoutubeSuggest *search = (YoutubeSuggest *)(object);

  if (search->async_http_session)
    g_object_unref(search->async_http_session);
  search->async_http_session = NULL;

  if (search->msg)
    g_object_unref(search->msg);

  G_OBJECT_CLASS (youtube_suggest_parent_class)->finalize(object);
}

static void
youtube_suggest_class_init(YoutubeSuggestClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

  gobject_class->finalize = youtube_suggest_finalize;

  youtube_suggest_signals[SIGNAL_SUGGEST_FINISHED] =
        g_signal_new("search-finished",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

}

static void
youtube_suggest_init(YoutubeSuggest *search)
{
  trace("\n");

  search->async_http_session = NULL;
}
#endif

