/*
 * gtktube: a gtk frontend for youtube like services
 * Copyright (C) 2009 Luc Saillard <luc@saillard.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gtktube.h"
#include "youtube_video.h"
#include "gtk_utils.h"
#include "utils.h"
#include "marshal.h"


#define DEBUG_XML 0



enum
{
  SIGNAL_ERROR,
  SIGNAL_STREAM_URL_READY,
  SIGNAL_LAST
};

static gint youtube_video_signals[SIGNAL_LAST];

struct youtube_cb
{
  int format;
  VideoGotUrlFunc cb;
  gpointer user_data;
};


/*
 *
 * fmt22 = HD[1280x720]
 * fmt35 = HQ[640x380]
 * fmt18 = mp4[480x360]
 * fmt34 = flv[320x180]
 * fmt17 = 3gp[176x144]
 * 
 * fmt5 seems to be equal to fmt34
 *
 * Example:
 *   fmt_map => "22/2000000/9/0/115,18/512000/9/0/115,34/0/9/0/115,5/0/7/0/0"
 *      ==> fmt22, fmt18, fmt34, fmt5 ???
 *
 */
static int
youtube_parse_fmt_map(YoutubeVideo *video, const char *raw_fmt)
{
  char **fmts;
  unsigned int i;
  unsigned int fmt, dummy;
  int fmtidx = 0;
  int best_fmt = 0;

  fmts = g_strsplit(raw_fmt, ",", -1);
  if (fmts == NULL)
    return 18;

  for (i=0; fmts[i] && fmtidx<G_N_ELEMENTS(video->sizes); i++)
   {
     if (sscanf(fmts[i], "%u/%u/%u/%u/%u", &fmt, &dummy, &dummy, &dummy, &dummy) != 5)
      {
	g_error("Format %s not supported\n", fmts[i]);
	continue;
      }

     if (best_fmt == 0)
       best_fmt = fmt;

     switch (fmt)
      {
       case 22:
       case 35:
       case 18:
       case 34:
       case 17:
	 video->sizes[fmtidx++] = fmt;
	 break;

       case 5: /* Ignoring fmt seems to be the same of 34 */
	 continue;

       default:
	 g_error("fmt \"%d\" is not supported, please report it to the developper\n", fmt);
	 continue;
      }
   }

  video->sizes[fmtidx] = 0;

  g_strfreev(fmts);

  return best_fmt;
}

static const char *
youtube_fmt_id_to_string(int fmt_id)
{
  switch (fmt_id)
   {
    case 22:
      return "1280x720";

    case 35:
      return "640x380";

    case 18:
      return "480x360";

    case 34:
      return "320x180";

    case 17:
      return "176x144";

    default:
      return "unknown";
   }
}


static void
youtube_get_video_info_async_callback(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
  YoutubeVideo *video = user_data;
  const char *data = msg->response_body->data;
  char *p;
  GHashTable *resp = NULL;
  const char *status, *reason, *video_id, *token, *raw_fmt;
  int fmt_selected;
  unsigned int i;

  if (ENABLE_DEBUG)
   {
     char *uri = soup_uri_to_string(soup_message_get_uri(msg), FALSE);
     trace("response from %s\n", uri);
     g_free(uri);
   }

  if (SOUP_STATUS_IS_TRANSPORT_ERROR(msg->status_code))
   {
exit_error:
     /* Report to the user an error */
     g_signal_emit(video, youtube_video_signals[SIGNAL_ERROR], 0);
     if (resp)
       g_hash_table_destroy(resp);
     return;
   } 

  resp = soup_form_decode(data);

  if (ENABLE_DEBUG)
   {
     GList *keys, *iter;

     keys = g_hash_table_get_keys(resp);

     for (iter = keys; iter; iter = iter->next)
      {
	char *k = (gchar *) iter->data;
	trace(" %s => \"%s\"\n", k, (char *)g_hash_table_lookup(resp, iter->data));
      }
   }


  status = g_hash_table_lookup(resp, "status");
  if (status == NULL)
    goto exit_error;

  if (strcmp(status, "ok"))
   {
     reason = g_hash_table_lookup(resp, "reason");
     if (reason == NULL)
       goto exit_error;

     /* Trim the reason to the first character we don't like */
     p = strchr(reason, '<');
     if (p)
       *p = 0;

     trace("Error: %s\n", reason);
     goto exit_error;
   }

  token = g_hash_table_lookup(resp, "token");
  if (token == NULL)
   {
     trace("We have not found the token in the webpage ...\n");
     goto exit_error;
   }

  video_id = g_hash_table_lookup(resp, "video_id");
  if (video_id == NULL)
   {
     trace("We have not found the video_id in the webpage ...\n");
     goto exit_error;
   }

  raw_fmt = g_hash_table_lookup(resp, "fmt_map");
  if (raw_fmt)
   {
     fmt_selected = youtube_parse_fmt_map(video, raw_fmt);
   }
  else
   {
     fmt_selected = 18;
     video->sizes[0] = 18;
     video->sizes[1] = 0;
   }

  if (video->fmt_selected > 0)
   {
     for (i=0; video->sizes[i]; i++)
      {
	if (video->fmt_selected == video->sizes[i])
	  break;
      }
     if (video->sizes[i] == 0)
       video->fmt_selected = 0;	/* the format selected is not available, unset it */
   }

  if (video->fmt_selected == 0)
    video->fmt_selected = fmt_selected;

  trace("token is \"%s\"\n", token);
  trace("video_id is \"%s\"\n", video_id);
  trace("fmt is \"%d\"\n", fmt_selected);

  if (video->token)
    g_free(video->token);
  video->token = g_strdup(token);

  if (video->video_id)
    g_free(video->video_id);
  video->video_id = g_strdup(video_id);

  g_hash_table_destroy(resp);

  g_signal_emit(video, youtube_video_signals[SIGNAL_STREAM_URL_READY], 0);
}



/*
 *
 * 
 *
 *
 */





/*
 *
 * Public interface
 *
 *
 */




YoutubeVideo *
youtube_video_new(void)
{
  YoutubeVideo *video = g_object_new(YOUTUBE_VIDEO_TYPE, NULL);

  return video;
}

void
youtube_set_video_size(YoutubeVideo *video, int width, int height)
{



}


/*
 * You can only call this function when the signal "stream-url-ready"
 * 
 * You must free the list
 */
GList *
youtube_get_video_sizes(YoutubeVideo *video)
{
  GList *list = NULL;
  unsigned int i;

  g_return_val_if_fail(video->token != NULL, NULL);

  for (i=0; video->sizes[i]; i++)
   {
     list = g_list_append(list, g_strdup(youtube_fmt_id_to_string(video->sizes[i])));
   }

  return list;
}

gboolean
youtube_video_set_video_url(YoutubeVideo *video, const char *url)
{
  char *get_video_info_url;
  const char *id;
  SoupMessage *msg;

  trace("\n");

  if (strncmp(url, "http://www.youtube.com/watch?v=", strlen("http://www.youtube.com/watch?v=")) != 0)
    return FALSE;

  id = url+strlen("http://www.youtube.com/watch?v=");

  get_video_info_url = g_strdup_printf("http://www.youtube.com/get_video_info?video_id=%s&el=embedded&ps=default&eurl=", id);

  msg = soup_message_new("GET", get_video_info_url);

  g_free(get_video_info_url);

  if (msg == NULL)
    return FALSE;

  if (video->http_session == NULL)
    video->http_session = soup_session_async_new();

  soup_session_queue_message(video->http_session, msg, youtube_get_video_info_async_callback, video);

  return TRUE;
}

char *
youtube_video_get_stream_video_url(YoutubeVideo *video)
{
  char *media_url;

  g_return_val_if_fail(video->token != NULL, NULL);
  g_return_val_if_fail(video->video_id != NULL, NULL);

  media_url = g_strdup_printf("http://www.youtube.com/get_video?video_id=%s&t=%s&eurl=&el=embedded&ps=default&fmt=%u",
     			      video->video_id, video->token, video->fmt_selected);


  return media_url;
}


/***********************************************************************
 *
 *
 *
 *
 * GObject things
 *
 *
 *
 *
 **********************************************************************/
G_DEFINE_TYPE(YoutubeVideo, youtube_video, G_TYPE_OBJECT)

static void
youtube_video_finalize(GObject *object)
{
  YoutubeVideo *video = (YoutubeVideo *)(object);

  if (video->token)
    g_free(video->token);
  video->token = NULL;

  if (video->video_id)
    g_free(video->video_id);
  video->video_id = NULL;


  G_OBJECT_CLASS (youtube_video_parent_class)->finalize(object);
}

static void
youtube_video_class_init(YoutubeVideoClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

  gobject_class->finalize = youtube_video_finalize;

  youtube_video_signals[SIGNAL_ERROR] =
        g_signal_new("error",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

  youtube_video_signals[SIGNAL_STREAM_URL_READY] =
        g_signal_new("stream-url-ready",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     g_cclosure_marshal_VOID__VOID,
                     G_TYPE_NONE, 0);

#if 0
  youtube_video_signals[SIGNAL_SEARCH_PROGRESS_STATUS] =
        g_signal_new("progress-status",
                     G_TYPE_FROM_CLASS(gobject_class),
                     G_SIGNAL_RUN_LAST,
                     0,
                     NULL, NULL,
                     gtktube_marshal_VOID__UINT64_UINT64,
                     G_TYPE_NONE,
		     2,
		     G_TYPE_UINT64,
		     G_TYPE_UINT64);

#endif
}

static void
youtube_video_init(YoutubeVideo *video)
{
  trace("\n");

}

